# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/control_surface.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 33300 bytes
import logging
import traceback
from collections import OrderedDict
from contextlib import contextmanager
from functools import partial
from itertools import chain
from pickle import dumps, loads

import Live
from future.utils import iteritems, itervalues

from ..base import (
    BooleanContext,
    EventObject,
    const,
    find_if,
    first,
    inject,
    lazy_attribute,
    liveobj_valid,
    old_hasattr,
    task,
)
from . import defaults, midi
from .control_element import OptimizedOwnershipHandler
from .device_bank_registry import DeviceBankRegistry
from .device_provider import DeviceProvider
from .elements import PhysicalDisplayElement
from .input_control_element import (
    MIDI_CC_TYPE,
    MIDI_NOTE_TYPE,
    MIDI_PB_TYPE,
    MIDI_SYSEX_TYPE,
    InputControlElement,
    ScriptForwarding,
)
from .message_scheduler import MessageScheduler
from .profile import profile

__all__ = ("SimpleControlSurface", "ControlSurface")
logger = logging.getLogger(__name__)
CS_LIST_KEY = "control_surfaces"


def publish_control_surface(control_surface):
    get_control_surfaces().append(control_surface)


def get_control_surfaces():
    if isinstance(__builtins__, dict):
        if CS_LIST_KEY not in __builtins__:
            __builtins__[CS_LIST_KEY] = []
        return __builtins__[CS_LIST_KEY]
    if not old_hasattr(__builtins__, CS_LIST_KEY):
        setattr(__builtins__, CS_LIST_KEY, [])
    return getattr(__builtins__, CS_LIST_KEY)


class SimpleControlSurface(EventObject):
    __events__ = ("received_midi", "disconnect")
    preferences_key = None
    handle_undo_steps = False

    def __init__(self, c_instance=None, *a, **k):
        (super().__init__)(*a, **k)
        self.canonical_parent = None
        publish_control_surface(self)
        self._c_instance = c_instance
        self._pad_translations = None
        self._components = []
        self._displays = []
        self._controls = []
        self._forwarding_long_identifier_registry = {}
        self._forwarding_registry = {}
        self._is_sending_scheduled_messages = BooleanContext()
        self._remaining_scheduled_messages = []
        self._task_group = task.TaskGroup(auto_kill=False)
        self._in_build_midi_map = BooleanContext()
        self._suppress_requests_counter = 0
        self._rebuild_requests_during_suppression = 0
        self._enabled = True
        self._in_component_guard = BooleanContext()
        self._accumulate_midi_messages = BooleanContext()
        self._midi_message_dict = {}
        self._midi_message_list = []
        self._midi_message_count = 0
        self.mxd_midi_scheduler = MessageScheduler(self._do_send_midi, self._task_group.add(task.TimedCallbackTask()))
        self._ownership_handler = OptimizedOwnershipHandler()
        self._control_surface_injector = inject(
            element_ownership_handler=(const(self._ownership_handler)),
            parent_task_group=(const(self._task_group)),
            show_message=(const(self.show_message)),
            register_component=(const(self._register_component)),
            register_control=(const(self._register_control)),
            request_rebuild_midi_map=(const(self.request_rebuild_midi_map)),
            set_pad_translations=(const(self.set_pad_translations)),
            send_midi=(const(self._send_midi)),
            song=(const(self.song)),
            set_session_highlight=(const(self._c_instance.set_session_highlight)),
        ).everywhere()

    @property
    def components(self):
        return tuple(filter(lambda comp: not comp.is_private, self._components))

    @property
    def root_components(self):
        return tuple(filter(lambda comp: comp.is_root and not comp.is_private, self._components))

    @property
    def controls(self):
        return self._controls

    def _get_tasks(self):
        return self._task_group

    _tasks = property(_get_tasks)

    @property
    def application(self):
        return Live.Application.get_application()

    @property
    def song(self):
        return self._c_instance.song()

    def disconnect(self):
        self._pre_serialize()
        self.notify_disconnect(self)
        self._disconnect_and_unregister_all_components()
        with self.component_guard():
            for control in self._controls:
                control.disconnect()
                control.canonical_parent = None

        self._forwarding_registry = None
        self._controls = None
        self._displays = None
        self._pad_translations = None
        cs_list = get_control_surfaces()
        if self in cs_list:
            cs_list.remove(self)
        self._task_group.clear()
        super().disconnect()

    def can_lock_to_devices(self):
        return False

    def suggest_map_mode(self, cc_no, channel):
        suggested_map_mode = -1
        for control in self._controls:
            if isinstance(control, InputControlElement):
                if control.message_type() == MIDI_CC_TYPE:
                    if control.message_identifier() == cc_no:
                        if control.message_channel() == channel:
                            suggested_map_mode = control.message_map_mode()
                            break

        return suggested_map_mode

    def supports_pad_translation(self):
        return self._pad_translations is not None

    def show_message(self, message):
        self._c_instance.show_message(message)

    def connect_script_instances(self, instanciated_scripts):
        pass

    def request_rebuild_midi_map(self):
        if self._suppress_requests_counter > 0:
            self._rebuild_requests_during_suppression += 1
        else:
            self._c_instance.request_rebuild_midi_map()

    def build_midi_map(self, midi_map_handle):
        with self._in_build_midi_map():
            self._forwarding_registry.clear()
            self._forwarding_long_identifier_registry.clear()
            for control in self._controls:
                if isinstance(control, InputControlElement):
                    control.install_connections(
                        self._translate_message,
                        partial(self._install_mapping, midi_map_handle),
                        partial(self._install_forwarding, midi_map_handle),
                    )

            if self._pad_translations is not None:
                self._c_instance.set_pad_translation(self._pad_translations)

    def port_settings_changed(self):
        self.refresh_state()

    def refresh_state(self):
        self.update()

    def update(self):
        with self.component_guard():
            for control in self._controls:
                control.clear_send_cache()

            for component in self._components:
                component.update()

    @profile
    def update_display(self):
        with self.component_guard():
            self.update_display_hook()
            with self._is_sending_scheduled_messages():
                self._task_group.update(defaults.TIMER_DELAY)

    def update_display_hook(self):
        pass

    @profile
    def receive_midi(self, midi_bytes):
        with self.component_guard():
            self._do_receive_midi(midi_bytes)

    @profile
    def receive_midi_chunk(self, midi_chunk):
        with self.component_guard():
            self._do_receive_midi_chunk(midi_chunk)

    @staticmethod
    def _receive_midi_data(recipient, data):
        recipient.receive_value(data)

    def _do_receive_midi(self, midi_bytes):
        (self.notify_received_midi)(*midi_bytes)
        self.mxd_midi_scheduler.handle_message(midi_bytes)
        self.process_midi_bytes(midi_bytes, self._receive_midi_data)

    @staticmethod
    def _merge_midi_data(recipient, data, midi_data):
        if recipient.allow_receiving_chunks:
            _, values = midi_data.setdefault(recipient, (recipient, []))
            values.append(data)
        else:
            midi_data[object()] = (recipient, data)

    def _do_receive_midi_chunk(self, midi_chunk):
        midi_data_for_recipient = OrderedDict()
        for midi_bytes in midi_chunk:
            (self.notify_received_midi)(*midi_bytes)
            self.mxd_midi_scheduler.handle_message(midi_bytes)
            self.process_midi_bytes(midi_bytes, partial((self._merge_midi_data), midi_data=midi_data_for_recipient))

        for recipient, data in itervalues(midi_data_for_recipient):
            if recipient.allow_receiving_chunks:
                recipient.receive_chunk(tuple(data))
            else:
                recipient.receive_value(data)

    def process_midi_bytes(self, midi_bytes, midi_processor):
        if midi.is_sysex(midi_bytes):
            result = self.get_registry_entry_for_sysex_midi_message(midi_bytes)
            if result is not None:
                identifier, recipient = result
                midi_processor(recipient, midi_bytes[len(identifier) : -1])
            else:
                if self.received_midi_listener_count() == 0:
                    logger.warning("Got unknown sysex message: " + midi.pretty_print_bytes(midi_bytes))
        else:
            recipient = self.get_recipient_for_nonsysex_midi_message(midi_bytes)
            if recipient is not None:
                midi_processor(recipient, midi.extract_value(midi_bytes))
            else:
                if self.received_midi_listener_count() == 0:
                    logger.warning("Got unknown message: " + midi.pretty_print_bytes(midi_bytes))

    def get_recipient_for_nonsysex_midi_message(self, midi_bytes):
        forwarding_key = midi_bytes[: 1 if midi.is_pitchbend(midi_bytes) else 2]
        if forwarding_key in self._forwarding_registry:
            return self._forwarding_registry[forwarding_key]
        return None

    def get_registry_entry_for_sysex_midi_message(self, midi_bytes):
        return find_if(
            lambda identifier_recipient: midi_bytes[: len(identifier_recipient[0])] == identifier_recipient[0],
            iteritems(self._forwarding_long_identifier_registry),
        )

    @contextmanager
    def suppressing_rebuild_requests(self):
        try:
            self._set_suppress_rebuild_requests(True)
            yield
        finally:
            self._set_suppress_rebuild_requests(False)

    def _set_suppress_rebuild_requests(self, suppress_requests):
        if suppress_requests:
            self._suppress_requests_counter += 1
        else:
            self._suppress_requests_counter -= 1
            if self._suppress_requests_counter == 0:
                if self._rebuild_requests_during_suppression > 0:
                    self.request_rebuild_midi_map()
                    self._rebuild_requests_during_suppression = 0

    def set_pad_translations(self, pad_translations):
        def check_translation(translation):
            return True

        self._pad_translations = pad_translations

    def set_enabled(self, enable):
        bool_enable = bool(enable)
        if self._enabled != bool_enable:
            with self.component_guard():
                self._enabled = bool_enable
                root_components = self.root_components
                components = root_components if len(root_components) > 0 else self._components
                for component in components:
                    component._set_enabled_recursive(bool_enable)

    def schedule_message(self, delay_in_ticks, callback, parameter=None):
        if not self._is_sending_scheduled_messages:
            delay_in_ticks -= 1
        message_reference = [None]

        def message(delta):
            if parameter:
                callback(parameter)
            else:
                callback()
            self._remaining_scheduled_messages.remove(message_reference)

        message_reference[0] = message
        self._remaining_scheduled_messages.append(message_reference)
        if delay_in_ticks:
            self._task_group.add(task.sequence(task.delay(delay_in_ticks), message))
        else:
            self._task_group.add(message)

    def set_feedback_channels(self, channels):
        self._c_instance.set_feedback_channels(channels)

    def set_controlled_track(self, track):
        self._c_instance.set_controlled_track(track)

    def release_controlled_track(self):
        self._c_instance.release_controlled_track()

    def _register_control(self, control):
        self._controls.append(control)
        control.canonical_parent = self
        if isinstance(control, PhysicalDisplayElement):
            self._displays.append(control)
        if old_hasattr(control, "_is_resource_based"):
            control._is_resource_based = True

    def _register_component(self, component):
        self._components.append(component)
        component.canonical_parent = self

    def _disconnect_and_unregister_all_components(self):
        with self.component_guard():
            for component in self._components:
                component.canonical_parent = None
                component.disconnect()

        self._components = []

    @contextmanager
    def component_guard(self):
        if not self._in_component_guard:
            with self._in_component_guard():
                with self._component_guard():
                    yield
        else:
            yield

    @contextmanager
    def _component_guard(self):
        with self._control_surface_injector:
            with self.suppressing_rebuild_requests():
                with self.accumulating_midi_messages():
                    yield
                    self._ownership_handler.commit_ownership_changes()

    @profile
    def call_listeners(self, listeners):
        with self.component_guard():
            for listener in filter(liveobj_valid, listeners):
                listener()

    @contextmanager
    def accumulating_midi_messages(self):
        with self._accumulate_midi_messages():
            try:
                yield
            finally:
                self._flush_midi_messages()

    def get_control_by_name(self, control_name):
        return find_if(lambda c: c.name == control_name, self._controls)

    def get_component_by_name(self, component_name):
        return find_if(lambda c: c.name == component_name, self.components)

    def _send_midi(self, midi_event_bytes, optimized=True):
        if self._accumulate_midi_messages:
            sysex_status_byte = 240
            entry = (self._midi_message_count, midi_event_bytes)
            if optimized and midi_event_bytes[0] != sysex_status_byte:
                key = (midi_event_bytes[0], midi_event_bytes[1])
                self._midi_message_dict[key] = entry
            else:
                self._midi_message_list.append(entry)
            self._midi_message_count += 1
        else:
            self._do_send_midi(midi_event_bytes)
        return True

    def _flush_midi_messages(self):
        sorted_messages = sorted((chain(self._midi_message_list, itervalues(self._midi_message_dict))), key=first)
        for _, message in sorted_messages:
            self._do_send_midi(message)

        self._midi_message_dict.clear()
        self._midi_message_list[:] = []
        self._midi_message_count = 0

    def _do_send_midi(self, midi_event_bytes):
        try:
            self._c_instance.send_midi(midi_event_bytes)
        except Exception:
            logger.error("Error while sending midi message %s", str(midi_event_bytes))
            traceback.print_exc()
            return False
        else:
            return True

    def _install_mapping(self, midi_map_handle, control, parameter, feedback_delay, feedback_map):
        success = False
        feedback_rule = None
        if control.message_type() == MIDI_NOTE_TYPE:
            feedback_rule = Live.MidiMap.NoteFeedbackRule()
            feedback_rule.note_no = control.message_identifier()
            feedback_rule.vel_map = feedback_map
        else:
            if control.message_type() == MIDI_CC_TYPE:
                feedback_rule = Live.MidiMap.CCFeedbackRule()
                feedback_rule.cc_no = control.message_identifier()
                feedback_rule.cc_value_map = feedback_map
            else:
                if control.message_type() == MIDI_PB_TYPE:
                    feedback_rule = Live.MidiMap.PitchBendFeedbackRule()
                    feedback_rule.value_pair_map = feedback_map
        feedback_rule.channel = control.message_channel()
        feedback_rule.delay_in_ms = feedback_delay
        feedback_rule.enabled = control.is_feedback_enabled
        if control.message_type() == MIDI_NOTE_TYPE:
            success = Live.MidiMap.map_midi_note_with_feedback_map(
                midi_map_handle,
                parameter,
                control.message_channel(),
                control.message_identifier(),
                feedback_rule,
            )
        else:
            if control.message_type() == MIDI_CC_TYPE:
                success = Live.MidiMap.map_midi_cc_with_feedback_map(
                    midi_map_handle,
                    parameter,
                    control.message_channel(),
                    control.message_identifier(),
                    control.message_map_mode(),
                    feedback_rule,
                    not control.needs_takeover(),
                    control.mapping_sensitivity,
                )
            else:
                if control.message_type() == MIDI_PB_TYPE:
                    success = Live.MidiMap.map_midi_pitchbend_with_feedback_map(
                        midi_map_handle,
                        parameter,
                        control.message_channel(),
                        feedback_rule,
                        not control.needs_takeover(),
                    )
        if success:
            Live.MidiMap.send_feedback_for_parameter(midi_map_handle, parameter)
        return success

    def _install_forwarding(self, midi_map_handle, control, forwarding_type=ScriptForwarding.exclusive):
        success = False
        should_consume_event = forwarding_type == ScriptForwarding.exclusive
        if control.message_type() == MIDI_NOTE_TYPE:
            success = Live.MidiMap.forward_midi_note(
                self._c_instance.handle(),
                midi_map_handle,
                control.message_channel(),
                control.message_identifier(),
                should_consume_event,
            )
        else:
            if control.message_type() == MIDI_CC_TYPE:
                success = Live.MidiMap.forward_midi_cc(
                    self._c_instance.handle(),
                    midi_map_handle,
                    control.message_channel(),
                    control.message_identifier(),
                    should_consume_event,
                )
            else:
                if control.message_type() == MIDI_PB_TYPE:
                    success = Live.MidiMap.forward_midi_pitchbend(
                        self._c_instance.handle(),
                        midi_map_handle,
                        control.message_channel(),
                    )
                else:
                    success = True
        if success:
            forwarding_keys = control.identifier_bytes()
            for key in forwarding_keys:
                registry = (
                    self._forwarding_registry
                    if control.message_type() != MIDI_SYSEX_TYPE
                    else self._forwarding_long_identifier_registry
                )
                registry[key] = control

        return success

    def _translate_message(self, type, from_identifier, from_channel, to_identifier, to_channel):
        if type == MIDI_CC_TYPE:
            self._c_instance.set_cc_translation(from_identifier, from_channel, to_identifier, to_channel)
        else:
            if type == MIDI_NOTE_TYPE:
                self._c_instance.set_note_translation(from_identifier, from_channel, to_identifier, to_channel)
            else:
                pass

    @lazy_attribute
    def preferences(self):
        if self.preferences_key is None:
            raise RuntimeError("Trying to access preferences without providing a preference_key")
        preferences = self._c_instance.preferences(self.preferences_key)
        pref_dict = {}
        try:
            pref_dict = loads(bytes(preferences))
        except Exception as e:
            try:
                logger.error("Error while loading control surface preferences %s", str(e))
            finally:
                e = None
                del e

        preferences.set_serializer(
            lambda: dumps(pref_dict),
        )
        return pref_dict

    def _pre_serialize(self):
        if self.preferences_key is not None:
            preferences = self._c_instance.preferences(self.preferences_key)
            dump = dumps(self.preferences)
            preferences.set_serializer(
                lambda: dump,
            )


class ControlSurface(SimpleControlSurface):
    device_provider_class = DeviceProvider

    def __init__(self, *a, **k):
        (super().__init__)(*a, **k)
        self._device_provider = None
        self._device_bank_registry = None
        if self.device_provider_class:
            self._init_device_provider()
        self._device_support_injector = inject(
            device_provider=(const(self.device_provider)),
            device_bank_registry=(const(self._device_bank_registry)),
        ).everywhere()

    def _init_device_provider(self):
        self._device_provider = self.register_disconnectable(self.device_provider_class(song=(self.song)))
        self._device_bank_registry = self.register_disconnectable(DeviceBankRegistry())
        self._c_instance.update_locks()
        self._device_provider.update_device_selection()

    @property
    def device_provider(self):
        return self._device_provider

    def disconnect(self):
        self._device_provider = None
        self._device_bank_registry = None
        super().disconnect()

    def can_lock_to_devices(self):
        return True

    def lock_to_device(self, device):
        with self.component_guard():
            self._device_provider.lock_to_device(device)

    def unlock_from_device(self, device):
        with self.component_guard():
            self._device_provider.unlock_from_device()

    def restore_bank(self, bank_index):
        device = self._device_provider.device
        if self._device_provider.is_locked_to_device:
            if liveobj_valid(device):
                with self.component_guard():
                    self._device_bank_registry.set_device_bank(device, bank_index)

    def toggle_lock(self):
        self._c_instance.toggle_lock()

    @contextmanager
    def _component_guard(self):
        with super()._component_guard():
            with self._device_support_injector:
                yield



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/base/live_api_utils.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 1636 bytes
import contextlib


def liveobj_changed(obj, other):
    return obj != other or type(obj) != type(other)


def liveobj_valid(obj):
    return obj is not None


def is_parameter_bipolar(param):
    return param.min == -1 * param.max


def duplicate_clip_loop(clip):
    if not liveobj_valid(clip) or clip.is_midi_clip:
        with contextlib.suppress(RuntimeError):
            clip.duplicate_loop()


def move_current_song_time(song, delta, truncate_to_beat=True):
    new_time = max(0, song.current_song_time + delta)
    if truncate_to_beat:
        new_time = int(new_time)
    song.current_song_time = new_time
    if not song.is_playing:
        song.start_time = new_time



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/_MxDCore/__init__.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 2238 bytes
import sys
import warnings

from ableton.v2.base import old_hasattr

from .MxDCore import MxDCore as _MxDCore


def set_manager(manager):
    _MxDCore.instance = _MxDCore()
    _MxDCore.instance.set_manager(manager)


def disconnect():
    _MxDCore.instance.disconnect()
    del _MxDCore.instance


def execute_command(device_id, object_id, command, arguments):
    if old_hasattr(_MxDCore.instance, command):
        try:
            with warnings.catch_warnings(record=True) as caught_warnings:
                _MxDCore.instance.update_device_context(device_id, object_id)
                function = getattr(_MxDCore.instance, command)
                function(device_id, object_id, arguments)
                for warning in caught_warnings:
                    _MxDCore.instance._print_warning(device_id, object_id, str(warning.message))

        except:
            assert_reason = str(sys.exc_info()[1]) if sys.exc_info()[0].__name__ == "RuntimeError" else "Invalid syntax"
            _MxDCore.instance._print_error(device_id, object_id, assert_reason)

    else:
        _MxDCore.instance._print_error(device_id, object_id, "Unknown command: " + command)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/wavetable_decoration.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 11563 bytes
import Live

from ..base import EventObject, listens, liveobj_valid
from .decoration import LiveObjectDecorator, NotifyingList, PitchParameter, get_parameter_by_name
from .internal_parameter import EnumWrappingParameter, IntegerParameter

ParameterState = Live.DeviceParameter.ParameterState


class WavetableEnvelopeType(int):
    pass


WavetableEnvelopeType.amp = WavetableEnvelopeType(0)
WavetableEnvelopeType.env2 = WavetableEnvelopeType(1)
WavetableEnvelopeType.env3 = WavetableEnvelopeType(2)


class WavetableLfoType(int):
    pass


WavetableLfoType.one = WavetableLfoType(0)
WavetableLfoType.two = WavetableLfoType(1)


class WavetableEnvelopeViewType(int):
    pass


WavetableEnvelopeViewType.time = WavetableEnvelopeViewType(0)
WavetableEnvelopeViewType.slope = WavetableEnvelopeViewType(1)
WavetableEnvelopeViewType.value = WavetableEnvelopeViewType(2)


class WavetableOscillatorType(int):
    pass


WavetableOscillatorType.one = WavetableOscillatorType(0)
WavetableOscillatorType.two = WavetableOscillatorType(1)
WavetableOscillatorType.s = WavetableOscillatorType(2)
WavetableOscillatorType.mix = WavetableOscillatorType(3)


class WavetableFilterType(int):
    pass


WavetableFilterType.one = WavetableFilterType(0)
WavetableFilterType.two = WavetableFilterType(1)


class WavetableDeviceDecorator(LiveObjectDecorator, EventObject):
    MIN_UNISON_VOICE_COUNT = 2
    MAX_UNISON_VOICE_COUNT = 8
    available_effect_modes = ("None", "Fm", "Classic", "Modern")
    available_unison_modes = ("None", "Classic", "Shimmer", "Noise", "Phase Sync", "Position Spread", "Random Note")
    mono_off_on_values = ("Off", "On")
    poly_voices_values = ("2", "3", "4", "5", "6", "7", "8")
    available_filter_routings = ("Serial", "Parallel", "Split")

    def __init__(self, *a, **k):
        (super().__init__)(*a, **k)
        self._osc_types_provider = NotifyingList(
            available_values=["1", "2", "S", "Mix"],
            default_value=(WavetableOscillatorType.one),
        )
        self._filter_types_provider = NotifyingList(
            available_values=["1", "2"],
            default_value=(WavetableFilterType.one),
        )
        self._envelope_types_provider = NotifyingList(
            available_values=["Amp", "Env2", "Env3"],
            default_value=(WavetableEnvelopeType.amp),
        )
        self._lfo_types_provider = NotifyingList(
            available_values=["LFO1", "LFO2"],
            default_value=(WavetableLfoType.one),
        )
        self._amp_envelope_view_types_provider = NotifyingList(
            available_values=["Time", "Slope"],
            default_value=(WavetableEnvelopeViewType.time),
        )
        self._mod_envelope_view_types_provider = NotifyingList(
            available_values=["Time", "Slope", "Value"],
            default_value=(WavetableEnvelopeViewType.time),
        )
        self._additional_parameters = self._create_parameters()
        self.register_disconnectables(self._additional_parameters)
        self._osc_1_on_parameter = get_parameter_by_name(self, "Osc 1 On")
        self._osc_2_on_parameter = get_parameter_by_name(self, "Osc 2 On")
        self._WavetableDeviceDecorator__on_osc_1_on_value_changed.subject = self._osc_1_on_parameter
        self._WavetableDeviceDecorator__on_osc_1_on_value_changed()
        self._WavetableDeviceDecorator__on_osc_2_on_value_changed.subject = self._osc_2_on_parameter
        self._WavetableDeviceDecorator__on_osc_2_on_value_changed()

    @property
    def parameters(self):
        return tuple(self._live_object.parameters) + self._additional_parameters

    def _create_parameters(self):
        self.oscillator_switch = EnumWrappingParameter(
            name="Oscillator",
            parent=self,
            values_host=(self._osc_types_provider),
            index_property_host=(self._osc_types_provider),
            values_property="available_values",
            index_property="index",
            value_type=WavetableOscillatorType,
        )
        self.osc_1_pitch = PitchParameter(
            name="Osc 1 Pitch",
            parent=self,
            integer_value_host=(get_parameter_by_name(self, "Osc 1 Transp")),
            decimal_value_host=(get_parameter_by_name(self, "Osc 1 Detune")),
        )
        self.osc_2_pitch = PitchParameter(
            name="Osc 2 Pitch",
            parent=self,
            integer_value_host=(get_parameter_by_name(self, "Osc 2 Transp")),
            decimal_value_host=(get_parameter_by_name(self, "Osc 2 Detune")),
        )
        self.envelope_switch = EnumWrappingParameter(
            name="Envelopes",
            parent=self,
            values_host=(self._envelope_types_provider),
            index_property_host=(self._envelope_types_provider),
            values_property="available_values",
            index_property="index",
            value_type=WavetableEnvelopeType,
        )
        self.lfo_switch = EnumWrappingParameter(
            name="LFO",
            parent=self,
            values_host=(self._lfo_types_provider),
            index_property_host=(self._lfo_types_provider),
            values_property="available_values",
            index_property="index",
            value_type=WavetableLfoType,
        )
        self._osc_1_category_switch = EnumWrappingParameter(
            name="Osc 1 Category",
            parent=self,
            values_host=(self._live_object),
            index_property_host=self,
            values_property="oscillator_wavetable_categories",
            index_property="oscillator_1_wavetable_category",
        )
        self._osc_2_category_switch = EnumWrappingParameter(
            name="Osc 2 Category",
            parent=self,
            values_host=(self._live_object),
            index_property_host=self,
            values_property="oscillator_wavetable_categories",
            index_property="oscillator_2_wavetable_category",
        )
        self._osc_1_table_switch = EnumWrappingParameter(
            name="Osc 1 Table",
            parent=self,
            values_host=(self._live_object),
            index_property_host=self,
            values_property="oscillator_1_wavetables",
            index_property="oscillator_1_wavetable_index",
        )
        self._osc_2_table_switch = EnumWrappingParameter(
            name="Osc 2 Table",
            parent=self,
            values_host=(self._live_object),
            index_property_host=self,
            values_property="oscillator_2_wavetables",
            index_property="oscillator_2_wavetable_index",
        )
        self._osc_1_effect_type_switch = EnumWrappingParameter(
            name="Osc 1 Effect Type",
            parent=self,
            values_host=self,
            index_property_host=self,
            values_property="available_effect_modes",
            index_property="oscillator_1_effect_mode",
        )
        self._osc_2_effect_type_switch = EnumWrappingParameter(
            name="Osc 2 Effect Type",
            parent=self,
            values_host=self,
            index_property_host=self,
            values_property="available_effect_modes",
            index_property="oscillator_2_effect_mode",
        )
        return (
            EnumWrappingParameter(
                name="Filter",
                parent=self,
                values_host=self._filter_types_provider,
                index_property_host=self._filter_types_provider,
                values_property="available_values",
                index_property="index",
                value_type=WavetableFilterType,
            ),
            EnumWrappingParameter(
                name="Amp Env View",
                parent=self,
                values_host=self._amp_envelope_view_types_provider,
                index_property_host=self._amp_envelope_view_types_provider,
                values_property="available_values",
                index_property="index",
                value_type=WavetableEnvelopeViewType,
            ),
            EnumWrappingParameter(
                name="Mod Env View",
                parent=self,
                values_host=self._mod_envelope_view_types_provider,
                index_property_host=self._mod_envelope_view_types_provider,
                values_property="available_values",
                index_property="index",
                value_type=WavetableEnvelopeViewType,
            ),
            EnumWrappingParameter(
                name="Unison Mode",
                parent=self,
                values_host=self,
                index_property_host=self,
                values_property="available_unison_modes",
                index_property="unison_mode",
            ),
            IntegerParameter(
                name="Unison Voices",
                parent=self,
                integer_value_host=self._live_object,
                integer_value_property_name="unison_voice_count",
                min_value=self.MIN_UNISON_VOICE_COUNT,
                max_value=self.MAX_UNISON_VOICE_COUNT,
                show_as_quantized=True,
            ),
            EnumWrappingParameter(
                name="Mono On",
                parent=self,
                values_host=self,
                index_property_host=self,
                values_property="mono_off_on_values",
                index_property="mono_poly",
                to_index_conversion=lambda index: int(not index),
                from_index_conversion=lambda index: int(not index),
            ),
            EnumWrappingParameter(
                name="Poly Voices",
                parent=self,
                values_host=self,
                index_property_host=self,
                values_property="poly_voices_values",
                index_property="poly_voices",
            ),
            EnumWrappingParameter(
                name="Filter Routing",
                parent=self,
                values_host=self,
                index_property_host=self,
                values_property="available_filter_routings",
                index_property="filter_routing",
            ),
            self.oscillator_switch,
            self.osc_1_pitch,
            self.osc_2_pitch,
            self.envelope_switch,
            self.lfo_switch,
            self._osc_1_category_switch,
            self._osc_2_category_switch,
            self._osc_1_table_switch,
            self._osc_2_table_switch,
            self._osc_1_effect_type_switch,
            self._osc_2_effect_type_switch,
        )

    def _get_parameter_enabled_state(self, parameter):
        if parameter.value:
            return ParameterState.enabled
        return ParameterState.disabled

    @listens("value")
    def __on_osc_1_on_value_changed(self):
        if liveobj_valid(self._osc_1_on_parameter):
            state = self._get_parameter_enabled_state(self._osc_1_on_parameter)
            self._osc_1_category_switch.state = state
            self._osc_1_table_switch.state = state
            self._osc_1_effect_type_switch.state = state
            self.osc_1_pitch.state = state

    @listens("value")
    def __on_osc_2_on_value_changed(self):
        if liveobj_valid(self._osc_2_on_parameter):
            state = self._get_parameter_enabled_state(self._osc_2_on_parameter)
            self._osc_2_category_switch.state = state
            self._osc_2_table_switch.state = state
            self._osc_2_effect_type_switch.state = state
            self.osc_2_pitch.state = state



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/instrument_finder.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 3943 bytes
from itertools import chain

import Live

from ..base import depends, listenable_property, listens_group
from ..live import liveobj_changed
from . import Component, find_instrument_devices, find_instrument_meeting_requirement
from .display import Renderable


def find_drum_group_device(track_or_chain):
    def requirement(instrument):
        return instrument.can_have_drum_pads

    return find_instrument_meeting_requirement(requirement, track_or_chain)


def find_sliced_simpler(track_or_chain):
    def requirement(instrument):
        return getattr(instrument, "playback_mode", None) == Live.SimplerDevice.PlaybackMode.slicing

    return find_instrument_meeting_requirement(requirement, track_or_chain)


class InstrumentFinderComponent(Component, Renderable):
    __events__ = ("instrument",)
    drum_group = listenable_property.managed(None)
    sliced_simpler = listenable_property.managed(None)
    any_instrument = listenable_property.managed(None)

    @depends(target_track=None)
    def __init__(self, name="Instrument_Finder", target_track=None, *a, **k):
        (super().__init__)(a, name=name, **k)
        self._target_track = target_track
        self.register_slot(target_track, self.update, "target_track")
        self.update()

    @listens_group("devices")
    def __on_devices_changed(self, _):
        self.update()

    @listens_group("chains")
    def __on_chains_changed(self, _):
        self.update()

    @listens_group("playback_mode")
    def __on_slicing_changed(self, _):
        self.update()

    def update(self):
        super().update()
        if self.is_enabled():
            self._update_listeners()
            self._update_instruments()

    def _update_listeners(self):
        track = self._target_track.target_track
        devices = list(find_instrument_devices(track))
        racks = list(filter(lambda d: d.can_have_chains, devices))
        simplers = list(filter(lambda d: hasattr(d, "playback_mode"), devices))
        chains = list(chain([track], *[d.chains for d in racks]))
        self._InstrumentFinderComponent__on_chains_changed.replace_subjects(racks)
        self._InstrumentFinderComponent__on_devices_changed.replace_subjects(chains)
        self._InstrumentFinderComponent__on_slicing_changed.replace_subjects(simplers)

    def _update_instruments(self):
        track = self._target_track.target_track
        do_notify = False
        drum_group = find_drum_group_device(track)
        if liveobj_changed(drum_group, self.drum_group):
            self.drum_group = drum_group
            do_notify = True
        sliced_simpler = find_sliced_simpler(track)
        if liveobj_changed(sliced_simpler, self.sliced_simpler):
            self.sliced_simpler = sliced_simpler
            do_notify = True
        any_instrument = list(find_instrument_devices(track))
        any_instrument = any_instrument[0] if any_instrument else None
        if liveobj_changed(any_instrument, self.any_instrument):
            self.any_instrument = any_instrument
            do_notify = True
        if do_notify:
            self.notify_instrument()



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/device_decorator_factory.py
# Compiled at: 2023-12-21 15:35:35
# Size of source mod 2**32: 1474 bytes
from ableton.v2.base import liveobj_valid
from ableton.v2.control_surface import DecoratorFactory

from .delay_decoration import DelayDeviceDecorator
from .drift_decoration import DriftDeviceDecorator
from .roar_decoration import RoarDeviceDecorator
from .simpler_decoration import SimplerDeviceDecorator
from .wavetable_decoration import WavetableDeviceDecorator


class DeviceDecoratorFactory(DecoratorFactory):
    DECORATOR_CLASSES = {
        "Delay": DelayDeviceDecorator,
        "Drift": DriftDeviceDecorator,
        "Roar": RoarDeviceDecorator,
        "OriginalSimpler": SimplerDeviceDecorator,
        "InstrumentVector": WavetableDeviceDecorator,
    }

    @classmethod
    def generate_decorated_device(cls, device, additional_properties=None, song=None, *a, **k):
        if additional_properties is None:
            additional_properties = {}
        return (cls.DECORATOR_CLASSES[device.class_name])(
            a,
            live_object=device,
            additional_properties=additional_properties,
            **k,
        )

    @classmethod
    def _should_be_decorated(cls, device):
        return liveobj_valid(device) and device.class_name in cls.DECORATOR_CLASSES

    def _get_decorated_object(self, device, additional_properties, song=None, *a, **k):
        return (self.generate_decorated_device)(device, *a, additional_properties=additional_properties, **k)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/components/view_toggle.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 2919 bytes
from ...base import MultiSlot
from .. import Component
from ..controls import ToggleButtonControl


class ViewToggleComponent(Component):
    main_view_toggle_button = ToggleButtonControl(color="ViewToggle.SessionOff", on_color="ViewToggle.SessionOn")
    detail_view_toggle_button = ToggleButtonControl(color="ViewToggle.DetailOff", on_color="ViewToggle.DetailOn")
    clip_view_toggle_button = ToggleButtonControl(color="ViewToggle.ClipOff", on_color="ViewToggle.ClipOn")
    browser_view_toggle_button = ToggleButtonControl(color="ViewToggle.BrowserOff", on_color="ViewToggle.BrowserOn")

    def __init__(self, name="View_Toggle", *a, **k):
        (super().__init__)(a, name=name, **k)
        for view_name in ("Session", "Detail", "Detail/Clip", "Browser"):
            self.register_slot(
                MultiSlot(
                    subject=(self.application.view),
                    listener=(self._ViewToggleComponent__update_view_toggle_buttons),
                    event_name_list=("is_view_visible",),
                    extra_args=(view_name,),
                ),
            )

        self._ViewToggleComponent__update_view_toggle_buttons()

    @main_view_toggle_button.toggled
    def main_view_toggle_button(self, is_toggled, _):
        self._show_or_hide_view(is_toggled, "Session")

    @detail_view_toggle_button.toggled
    def detail_view_toggle_button(self, is_toggled, _):
        self._show_or_hide_view(is_toggled, "Detail")

    @clip_view_toggle_button.toggled
    def clip_view_toggle_button(self, is_toggled, _):
        self._show_or_hide_view(is_toggled, "Detail/Clip")

    @browser_view_toggle_button.toggled
    def browser_view_toggle_button(self, is_toggled, _):
        self._show_or_hide_view(is_toggled, "Browser")

    def _show_or_hide_view(self, show_view, view_name):
        if show_view:
            self.application.view.show_view(view_name)
        else:
            self.application.view.hide_view(view_name)

    def __update_view_toggle_buttons(self):
        view = self.application.view
        self.main_view_toggle_button.is_on = view.is_view_visible("Session")
        self.detail_view_toggle_button.is_on = view.is_view_visible("Detail")
        self.clip_view_toggle_button.is_on = view.is_view_visible("Detail/Clip")
        self.browser_view_toggle_button.is_on = view.is_view_visible("Browser")



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/_Framework/SessionComponent.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 26304 bytes
from itertools import count

import Live

from .CompoundComponent import CompoundComponent
from .SceneComponent import SceneComponent
from .ScrollComponent import ScrollComponent
from .SubjectSlot import subject_slot, subject_slot_group
from .Util import in_range, product


class SessionComponent(CompoundComponent):
    __subject_events__ = ("offset",)
    _linked_session_instances = []
    _minimal_track_offset = -1
    _minimal_scene_offset = -1
    _highlighting_callback = None
    _session_component_ends_initialisation = True
    scene_component_type = SceneComponent

    def __init__(self, num_tracks=0, num_scenes=0, auto_name=False, enable_skinning=False, *a, **k):
        (super().__init__)(*a, **k)
        self._track_offset = -1
        self._scene_offset = -1
        self._num_tracks = num_tracks
        self._num_scenes = num_scenes
        (
            self._vertical_banking,
            self._horizontal_banking,
            self._vertical_paginator,
            self._horizontal_paginator,
        ) = self.register_components(ScrollComponent(), ScrollComponent(), ScrollComponent(), ScrollComponent())
        self._vertical_banking.can_scroll_up = self._can_bank_up
        self._vertical_banking.can_scroll_down = self._can_bank_down
        self._vertical_banking.scroll_up = self._bank_up
        self._vertical_banking.scroll_down = self._bank_down
        self._horizontal_banking.can_scroll_up = self._can_bank_left
        self._horizontal_banking.can_scroll_down = self._can_bank_right
        self._horizontal_banking.scroll_up = self._bank_left
        self._horizontal_banking.scroll_down = self._bank_right
        self._vertical_paginator.can_scroll_up = self._can_scroll_page_up
        self._vertical_paginator.can_scroll_down = self._can_scroll_page_down
        self._vertical_paginator.scroll_up = self._scroll_page_up
        self._vertical_paginator.scroll_down = self._scroll_page_down
        self._horizontal_paginator.can_scroll_up = self._can_scroll_page_left
        self._horizontal_paginator.can_scroll_down = self._can_scroll_page_right
        self._horizontal_paginator.scroll_up = self._scroll_page_left
        self._horizontal_paginator.scroll_down = self._scroll_page_right
        self._page_left_button = None
        self._page_right_button = None
        self._stop_all_button = None
        self._next_scene_button = None
        self._prev_scene_button = None
        self._stop_track_clip_buttons = None
        self._stop_clip_triggered_value = 127
        self._stop_clip_value = None
        self._highlighting_callback = None
        self._show_highlight = num_tracks > 0 and num_scenes > 0
        self._mixer = None
        self._track_slots = self.register_slot_manager()
        self._selected_scene = self.register_component(self._create_scene())
        self._scenes = (self.register_components)(*[self._create_scene() for _ in range(num_scenes)])
        if self._session_component_ends_initialisation:
            self._end_initialisation()
        if auto_name:
            self._auto_name()
        if enable_skinning:
            self._enable_skinning()

    def _end_initialisation(self):
        self.on_selected_scene_changed()
        self.set_offsets(0, 0)

    def _create_scene(self):
        return self.scene_component_type(num_slots=(self._num_tracks), tracks_to_use_callback=(self.tracks_to_use))

    def disconnect(self):
        if self._is_linked():
            self._unlink()
        super(CompoundComponent, self).disconnect()

    def set_highlighting_callback(self, callback):
        if self._highlighting_callback != callback:
            self._highlighting_callback = callback
            self._do_show_highlight()

    def scene(self, index):
        return self._scenes[index]

    def selected_scene(self):
        return self._selected_scene

    def _enable_skinning(self):
        self.set_stop_clip_triggered_value("Session.StopClipTriggered")
        self.set_stop_clip_value("Session.StopClip")
        for scene_index in range(self._num_scenes):
            scene = self.scene(scene_index)
            scene.set_scene_value("Session.Scene")
            scene.set_no_scene_value("Session.NoScene")
            scene.set_triggered_value("Session.SceneTriggered")
            for track_index in range(self._num_tracks):
                clip_slot = scene.clip_slot(track_index)
                clip_slot.set_triggered_to_play_value("Session.ClipTriggeredPlay")
                clip_slot.set_triggered_to_record_value("Session.ClipTriggeredRecord")
                clip_slot.set_record_button_value("Session.RecordButton")
                clip_slot.set_stopped_value("Session.ClipStopped")
                clip_slot.set_started_value("Session.ClipStarted")
                clip_slot.set_recording_value("Session.ClipRecording")

    def _auto_name(self):
        self.name = "Session_Control"
        self.selected_scene().name = "Selected_Scene"
        for track_index in range(self._num_tracks):
            clip_slot = self.selected_scene().clip_slot(track_index)
            clip_slot.name = "Selected_Scene_Clip_Slot_%d" % track_index

        for scene_index in range(self._num_scenes):
            scene = self.scene(scene_index)
            scene.name = "Scene_%d" % scene_index
            for track_index in range(self._num_tracks):
                clip_slot = scene.clip_slot(track_index)
                clip_slot.name = "%d_Clip_Slot_%d" % (track_index, scene_index)

    def set_scene_bank_buttons(self, down_button, up_button):
        self.set_scene_bank_up_button(up_button)
        self.set_scene_bank_down_button(down_button)

    def set_scene_bank_up_button(self, button):
        self._bank_up_button = button
        self._vertical_banking.set_scroll_up_button(button)

    def set_scene_bank_down_button(self, button):
        self._bank_down_button = button
        self._vertical_banking.set_scroll_down_button(button)

    def set_track_bank_buttons(self, right_button, left_button):
        self.set_track_bank_left_button(left_button)
        self.set_track_bank_right_button(right_button)

    def set_track_bank_left_button(self, button):
        self._bank_left_button = button
        self._horizontal_banking.set_scroll_up_button(button)

    def set_track_bank_right_button(self, button):
        self._bank_right_button = button
        self._horizontal_banking.set_scroll_down_button(button)

    def set_page_up_button(self, page_up_button):
        self._vertical_paginator.set_scroll_up_button(page_up_button)

    def set_page_down_button(self, page_down_button):
        self._vertical_paginator.set_scroll_down_button(page_down_button)

    def set_page_left_button(self, page_left_button):
        self._page_left_button = page_left_button
        self._horizontal_paginator.set_scroll_up_button(page_left_button)

    def set_page_right_button(self, page_right_button):
        self._page_right_button = page_right_button
        self._horizontal_paginator.set_scroll_down_button(page_right_button)

    def _can_scroll_page_up(self):
        return self.scene_offset() > 0

    def _can_scroll_page_down(self):
        return self.scene_offset() < len(self.song().scenes) - self.height()

    def _scroll_page_up(self):
        height = self.height()
        track_offset = self.track_offset()
        scene_offset = self.scene_offset()
        if scene_offset > 0:
            new_scene_offset = scene_offset
            if scene_offset % height > 0:
                new_scene_offset -= scene_offset % height
            else:
                new_scene_offset = max(0, scene_offset - height)
            self.set_offsets(track_offset, new_scene_offset)

    def _scroll_page_down(self):
        height = self.height()
        track_offset = self.track_offset()
        scene_offset = self.scene_offset()
        new_scene_offset = scene_offset + height - scene_offset % height
        self.set_offsets(track_offset, new_scene_offset)

    def _can_scroll_page_left(self):
        return self.track_offset() > 0

    def _can_scroll_page_right(self):
        return self.track_offset() < len(self.tracks_to_use()) - self.width()

    def _scroll_page_left(self):
        width = self.width()
        track_offset = self.track_offset()
        scene_offset = self.scene_offset()
        if track_offset > 0:
            new_track_offset = track_offset
            if track_offset % width > 0:
                new_track_offset -= track_offset % width
            else:
                new_track_offset = max(0, track_offset - width)
            self.set_offsets(new_track_offset, scene_offset)

    def _scroll_page_right(self):
        width = self.width()
        track_offset = self.track_offset()
        scene_offset = self.scene_offset()
        new_track_offset = track_offset + width - track_offset % width
        self.set_offsets(new_track_offset, scene_offset)

    def set_stop_all_clips_button(self, button):
        self._stop_all_button = button
        self._on_stop_all_value.subject = button
        self._update_stop_all_clips_button()

    def set_stop_track_clip_buttons(self, buttons):
        self._stop_track_clip_buttons = buttons
        self._on_stop_track_value.replace_subjects(buttons or [])
        self._update_stop_track_clip_buttons()

    def set_stop_clip_triggered_value(self, value):
        self._stop_clip_triggered_value = value

    def set_stop_clip_value(self, value):
        self._stop_clip_value = value

    def set_select_buttons(self, next_button, prev_button):
        self.set_select_next_button(next_button)
        self.set_select_prev_button(prev_button)

    def set_select_next_button(self, next_button):
        self._next_scene_button = next_button
        self._on_next_scene_value.subject = next_button
        self._update_select_buttons()

    def set_select_prev_button(self, prev_button):
        self._prev_scene_button = prev_button
        self._on_prev_scene_value.subject = prev_button
        self._update_select_buttons()

    def set_clip_launch_buttons(self, buttons):
        if buttons:
            for button, (x, y) in buttons.iterbuttons():
                scene = self.scene(y)
                slot = scene.clip_slot(x)
                slot.set_launch_button(button)

        else:
            for x, y in product(range(self._num_tracks), range(self._num_scenes)):
                scene = self.scene(y)
                slot = scene.clip_slot(x)
                slot.set_launch_button(None)

    def set_scene_launch_buttons(self, buttons):
        if buttons:
            for button, (x, _) in buttons.iterbuttons():
                scene = self.scene(x)
                scene.set_launch_button(button)

        else:
            for x in range(self._num_scenes):
                scene = self.scene(x)
                scene.set_launch_button(None)

    def set_mixer(self, mixer):
        self._mixer = mixer
        if self._mixer is not None:
            self._mixer.set_track_offset(self.track_offset())

    def set_offsets(self, track_offset, scene_offset):
        track_increment = 0
        scene_increment = 0
        if self._is_linked():
            SessionComponent._perform_offset_change(
                track_offset - self._track_offset,
                scene_offset - self._scene_offset,
            )
        else:
            if len(self.tracks_to_use()) > track_offset:
                track_increment = track_offset - self._track_offset
            if len(self.song().scenes) > scene_offset:
                scene_increment = scene_offset - self._scene_offset
            self._change_offsets(track_increment, scene_increment)

    def set_show_highlight(self, show_highlight):
        if self._show_highlight != show_highlight:
            self._show_highlight = show_highlight
            self._do_show_highlight()

    def set_rgb_mode(self, color_palette, color_table, clip_slots_only=False):
        for y in range(self._num_scenes):
            scene = self.scene(y)
            if not clip_slots_only:
                scene.set_color_palette(color_palette)
                scene.set_color_table(color_table)
            else:
                for x in range(self._num_tracks):
                    slot = scene.clip_slot(x)
                    slot.set_clip_palette(color_palette)
                    slot.set_clip_rgb_table(color_table)

    def on_enabled_changed(self):
        self.update()
        self._do_show_highlight()

    def update(self):
        super().update()
        if self._allow_updates:
            self._update_select_buttons()
            self._update_stop_track_clip_buttons()
            self._update_stop_all_clips_button()
        else:
            self._update_requests += 1

    def _update_stop_track_clip_buttons(self):
        if self.is_enabled():
            for index in range(self._num_tracks):
                self._update_stop_clips_led(index)

    def on_scene_list_changed(self):
        if not self._update_scene_offset():
            self._reassign_scenes()

    def _snap_track_offset(self):
        return self._page_left_button or self._page_right_button

    def on_track_list_changed(self):
        num_tracks = len(self.tracks_to_use())
        new_track_offset = self.track_offset()
        if new_track_offset >= num_tracks:
            new_track_offset = num_tracks - 1
            new_track_offset -= new_track_offset % (self.width() if self._snap_track_offset() else 1)
        self._reassign_tracks()
        self.set_offsets(new_track_offset, self.scene_offset())

    def on_selected_scene_changed(self):
        self._update_scene_offset()
        if self._selected_scene is not None:
            self._selected_scene.set_scene(self.song().view.selected_scene)

    def width(self):
        return self._num_tracks

    def height(self):
        return len(self._scenes)

    def track_offset(self):
        return self._track_offset

    def scene_offset(self):
        return self._scene_offset

    def tracks_to_use(self):
        return self._mixer.tracks_to_use() if self._mixer is not None else self.song().visible_tracks

    @property
    def current_tracks(self):
        tracks_to_use = self.tracks_to_use()
        offset = self.track_offset()
        return tracks_to_use[offset : offset + self.width()]

    def _get_minimal_track_offset(self):
        if self._is_linked():
            return SessionComponent._minimal_track_offset
        return self.track_offset()

    def _get_minimal_scene_offset(self):
        if self._is_linked():
            return SessionComponent._minimal_scene_offset
        return self.scene_offset()

    def _can_bank_down(self):
        return len(self.song().scenes) > self._get_minimal_scene_offset() + 1

    def _can_bank_up(self):
        return self._get_minimal_scene_offset() > 0

    def _can_bank_right(self):
        return len(self.tracks_to_use()) > self._get_minimal_track_offset() + 1

    def _can_bank_left(self):
        return self._get_minimal_track_offset() > 0

    def _bank_up(self):
        return self.set_offsets(self.track_offset(), max(0, self.scene_offset() - 1))

    def _bank_down(self):
        return self.set_offsets(self.track_offset(), self.scene_offset() + 1)

    def _bank_right(self):
        return self.set_offsets(self.track_offset() + 1, self.scene_offset())

    def _bank_left(self):
        return self.set_offsets(max(self.track_offset() - 1, 0), self.scene_offset())

    def _update_stop_all_clips_button(self):
        button = self._stop_all_button
        if button:
            button.set_light(button.is_pressed())

    def _update_select_buttons(self):
        selected_scene = self.song().view.selected_scene
        if self._next_scene_button is not None:
            self._next_scene_button.set_light(selected_scene != self.song().scenes[-1])
        if self._prev_scene_button is not None:
            self._prev_scene_button.set_light(selected_scene != self.song().scenes[0])

    def _update_scene_offset(self):
        offset_corrected = False
        num_scenes = len(self.song().scenes)
        if self.scene_offset() >= num_scenes:
            self.set_offsets(self.track_offset(), num_scenes - 1)
            offset_corrected = True
        return offset_corrected

    def _change_offsets(self, track_increment, scene_increment):
        offsets_changed = track_increment != 0 or scene_increment != 0
        if offsets_changed:
            self._track_offset += track_increment
            self._scene_offset += scene_increment
            if self._mixer is not None:
                self._mixer.set_track_offset(self.track_offset())
            self._reassign_tracks()
            self._reassign_scenes()
            self.notify_offset()
            if self.width() > 0:
                if self.height() > 0:
                    self._do_show_highlight()

    def _reassign_scenes(self):
        scenes = self.song().scenes
        for index, scene in enumerate(self._scenes):
            scene_index = self._scene_offset + index
            if len(scenes) > scene_index:
                scene.set_scene(scenes[scene_index])
                scene.set_track_offset(self._track_offset)
            else:
                self._scenes[index].set_scene(None)

        if self._selected_scene is not None:
            self._selected_scene.set_track_offset(self._track_offset)
        self._vertical_banking.update()
        self._vertical_paginator.update()

    def _reassign_tracks(self):
        current_tracks = self.current_tracks
        self._on_fired_slot_index_changed.replace_subjects(current_tracks, count())
        self._on_playing_slot_index_changed.replace_subjects(current_tracks, count())
        self._horizontal_banking.update()
        self._horizontal_paginator.update()
        self._update_stop_all_clips_button()
        self._update_stop_track_clip_buttons()

    @subject_slot("value")
    def _on_stop_all_value(self, value):
        self._stop_all_value(value)

    def _stop_all_value(self, value):
        if self.is_enabled():
            if not (value != 0 or self._stop_all_button.is_momentary()):
                self.song().stop_all_clips()
            self._update_stop_all_clips_button()

    @subject_slot("value")
    def _on_next_scene_value(self, value):
        if self.is_enabled():
            if not (value != 0 or self._next_scene_button.is_momentary()):
                selected_scene = self.song().view.selected_scene
                all_scenes = self.song().scenes
                if selected_scene != all_scenes[-1]:
                    index = list(all_scenes).index(selected_scene)
                    self.song().view.selected_scene = all_scenes[index + 1]

    @subject_slot("value")
    def _on_prev_scene_value(self, value):
        if self.is_enabled():
            if not (value != 0 or self._prev_scene_button.is_momentary()):
                selected_scene = self.song().view.selected_scene
                all_scenes = self.song().scenes
                if selected_scene != all_scenes[0]:
                    index = list(all_scenes).index(selected_scene)
                    self.song().view.selected_scene = all_scenes[index - 1]

    @subject_slot_group("value")
    def _on_stop_track_value(self, value, button):
        if self.is_enabled():
            if not (value != 0 or button.is_momentary()):
                tracks = self.tracks_to_use()
                track_index = list(self._stop_track_clip_buttons).index(button) + self.track_offset()
                if in_range(track_index, 0, len(tracks)):
                    if tracks[track_index] in self.song().tracks:
                        tracks[track_index].stop_all_clips()

    def _do_show_highlight(self):
        if self._highlighting_callback is not None:
            return_tracks = self.song().return_tracks
            include_returns = len(return_tracks) > 0 and return_tracks[0] in self.tracks_to_use()
            if self._show_highlight:
                self._highlighting_callback(
                    self._track_offset,
                    self._scene_offset,
                    self.width(),
                    self.height(),
                    include_returns,
                )
            else:
                self._highlighting_callback(-1, -1, -1, -1, include_returns)

    @subject_slot_group("fired_slot_index")
    def _on_fired_slot_index_changed(self, controlled_track_index):
        self._update_stop_clips_led(controlled_track_index)

    @subject_slot_group("playing_slot_index")
    def _on_playing_slot_index_changed(self, controlled_track_index):
        self._update_stop_clips_led(controlled_track_index)

    def _update_stop_clips_led(self, index):
        tracks_to_use = self.tracks_to_use()
        track_index = index + self.track_offset()
        if self.is_enabled():
            if self._stop_track_clip_buttons is not None:
                if index < len(self._stop_track_clip_buttons):
                    button = self._stop_track_clip_buttons[index]
                    if button is not None:
                        value_to_send = None
                        if track_index < len(tracks_to_use):
                            if tracks_to_use[track_index].clip_slots:
                                track = tracks_to_use[track_index]
                                if track.fired_slot_index == -2:
                                    value_to_send = self._stop_clip_triggered_value
                                else:
                                    if track.playing_slot_index >= 0:
                                        value_to_send = self._stop_clip_value
                        if value_to_send is None:
                            button.turn_off()
                        else:
                            if in_range(value_to_send, 0, 128):
                                button.send_value(value_to_send)
                            else:
                                button.set_light(value_to_send)

    def _is_linked(self):
        return self in SessionComponent._linked_session_instances

    def _link(self):
        SessionComponent._linked_session_instances.append(self)

    def _unlink(self):
        SessionComponent._linked_session_instances.remove(self)

    @staticmethod
    def _perform_offset_change(track_increment, scene_increment):
        scenes = Live.Application.get_application().get_document().scenes
        instances_covering_session = 0
        found_negative_offset = False
        minimal_track_offset = -1
        minimal_scene_offset = -1
        for instance in SessionComponent._linked_session_instances:
            new_track_offset = instance.track_offset() + track_increment
            new_scene_offset = instance.scene_offset() + scene_increment
            if new_track_offset >= 0:
                if new_scene_offset >= 0:
                    if new_track_offset < len(instance.tracks_to_use()) and new_scene_offset < len(scenes):
                        instances_covering_session += 1
                        if minimal_track_offset < 0:
                            minimal_track_offset = new_track_offset
                        else:
                            minimal_track_offset = min(minimal_track_offset, new_track_offset)
                        if minimal_scene_offset < 0:
                            minimal_scene_offset = new_scene_offset
                        else:
                            minimal_scene_offset = min(minimal_scene_offset, new_scene_offset)
                    else:
                        found_negative_offset = True
                        break

        if not found_negative_offset:
            if instances_covering_session > 0:
                SessionComponent._minimal_track_offset = minimal_track_offset
                SessionComponent._minimal_scene_offset = minimal_scene_offset
                for instance in SessionComponent._linked_session_instances:
                    instance._change_offsets(track_increment, scene_increment)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/_Generic/SpecialMixerComponent.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 661 bytes
from _Framework import MixerComponent

from .SelectChanStripComponent import SelectChanStripComponent


class SpecialMixerComponent(MixerComponent):
    def _create_strip(self):
        return SelectChanStripComponent()

    def set_bank_up_button(self, button):
        self.set_bank_buttons(button, self._bank_down_button)

    def set_bank_down_button(self, button):
        self.set_bank_buttons(self._bank_up_button, button)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/_Framework/SceneComponent.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 7139 bytes
from .ClipSlotComponent import ClipSlotComponent, find_nearest_color
from .CompoundComponent import CompoundComponent
from .SubjectSlot import subject_slot
from .Util import in_range, nop


class SceneComponent(CompoundComponent):
    clip_slot_component_type = ClipSlotComponent

    def __init__(self, num_slots=0, tracks_to_use_callback=nop, *a, **k):
        (super().__init__)(*a, **k)
        self._scene = None
        self._clip_slots = []
        self._tracks_to_use_callback = tracks_to_use_callback
        self._color_palette = None
        self._color_table = None
        for _ in range(num_slots):
            new_slot = self._create_clip_slot()
            self._clip_slots.append(new_slot)
            self.register_components(new_slot)

        self._launch_button = None
        self._triggered_value = 127
        self._scene_value = None
        self._no_scene_value = None
        self._track_offset = 0
        self._select_button = None
        self._delete_button = None

    def on_track_list_changed(self):
        self.update()

    def set_scene(self, scene):
        if scene != self._scene or type(self._scene) != type(scene):
            self._scene = scene
            self._on_is_triggered_changed.subject = scene
            self._on_scene_color_changed.subject = scene
            self.update()

    def set_launch_button(self, button):
        if button != self._launch_button:
            self._launch_button = button
            self._launch_value.subject = button
            self.update()

    def set_select_button(self, button):
        self._select_button = button

    def set_delete_button(self, button):
        self._delete_button = button

    def set_track_offset(self, offset):
        if offset != self._track_offset:
            self._track_offset = offset
            self.update()

    def set_triggered_value(self, value):
        self._triggered_value = value

    def set_scene_value(self, value):
        self._scene_value = value

    def set_no_scene_value(self, value):
        self._no_scene_value = value

    def set_color_palette(self, palette):
        self._scene_value = None
        self._color_palette = palette

    def set_color_table(self, table):
        self._scene_value = None
        self._color_table = table

    def clip_slot(self, index):
        return self._clip_slots[index]

    def update(self):
        super().update()
        if self._allow_updates:
            if self._scene is not None and self.is_enabled():
                clip_index = self._track_offset
                tracks = self.song().tracks
                clip_slots = self._scene.clip_slots
                if self._track_offset > 0:
                    real_offset = 0
                    visible_tracks = 0
                    while visible_tracks < self._track_offset:
                        if len(tracks) > real_offset:
                            if tracks[real_offset].is_visible:
                                visible_tracks += 1
                            else:
                                real_offset += 1

                    clip_index = real_offset
                for slot in self._clip_slots:
                    while len(tracks) > clip_index:
                        if not tracks[clip_index].is_visible:
                            clip_index += 1

                    if len(clip_slots) > clip_index:
                        slot.set_clip_slot(clip_slots[clip_index])
                    else:
                        slot.set_clip_slot(None)
                    clip_index += 1

            else:
                for slot in self._clip_slots:
                    slot.set_clip_slot(None)

            self._update_launch_button()
        else:
            self._update_requests += 1

    @subject_slot("value")
    def _launch_value(self, value):
        if self.is_enabled():
            if self._select_button and self._select_button.is_pressed() and value:
                self._do_select_scene(self._scene)
            else:
                if self._scene is not None:
                    if self._delete_button and self._delete_button.is_pressed() and value:
                        self._do_delete_scene(self._scene)
                    else:
                        self._do_launch_scene(value)

    def _do_select_scene(self, scene_for_overrides):
        if self._scene is not None:
            view = self.song().view
            if view.selected_scene != self._scene:
                view.selected_scene = self._scene

    def _do_delete_scene(self, scene_for_overrides):
        try:
            if self._scene:
                song = self.song()
                song.delete_scene(list(song.scenes).index(self._scene))
        except RuntimeError:
            pass

    def _do_launch_scene(self, value):
        launched = False
        if self._launch_button.is_momentary():
            self._scene.set_fire_button_state(value != 0)
            launched = value != 0
        else:
            if value != 0:
                self._scene.fire()
                launched = True
        if launched:
            if self.song().select_on_launch:
                self.song().view.selected_scene = self._scene

    @subject_slot("is_triggered")
    def _on_is_triggered_changed(self):
        self._update_launch_button()

    @subject_slot("color")
    def _on_scene_color_changed(self):
        self._update_launch_button()

    def _color_value(self, color):
        value = None
        if self._color_palette:
            value = self._color_palette.get(color, None)
        if value is None:
            if self._color_table:
                value = find_nearest_color(self._color_table, color)
        return value

    def _update_launch_button(self):
        if self.is_enabled():
            if self._launch_button is not None:
                value_to_send = self._no_scene_value
                if self._scene:
                    if self._scene.is_triggered:
                        value_to_send = self._triggered_value
                    else:
                        if self._scene_value is not None:
                            value_to_send = self._scene_value
                        else:
                            value_to_send = self._color_value(self._scene.color)
                if value_to_send is None:
                    self._launch_button.turn_off()
                else:
                    if in_range(value_to_send, 0, 128):
                        self._launch_button.send_value(value_to_send)
                    else:
                        self._launch_button.set_light(value_to_send)

    def _create_clip_slot(self):
        return self.clip_slot_component_type()



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/components/session_ring.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 9227 bytes
from functools import partial

from ...base import const, depends, index_if, listenable_property, listens, nop
from .. import Component
from ..display import Renderable


class SessionRingModel:
    def __init__(self, num_tracks, num_scenes, include_returns, set_session_highlight=nop):
        self.num_tracks = num_tracks
        self.num_scenes = num_scenes
        self.include_returns = include_returns
        self.track_offset = 0
        self.scene_offset = 0
        self.callback = set_session_highlight

    def update_highlight(self):
        self.callback(self.track_offset, self.scene_offset, self.num_tracks, self.num_scenes, self.include_returns)

    def hide_highlight(self):
        self.callback(-1, -1, -1, -1, False)

    def move(self, tracks, scenes):
        self.track_offset += tracks
        self.scene_offset += scenes


class SessionRingComponent(Component, Renderable):
    @depends(set_session_highlight=(const(nop)))
    def __init__(
        self,
        name="Session_Ring",
        num_tracks=0,
        num_scenes=0,
        include_returns=False,
        include_master=False,
        right_align_non_player_tracks=False,
        tracks_to_use=None,
        snap_track_offset=False,
        set_session_highlight=nop,
        is_private=False,
        *a,
        **k,
    ):
        (super().__init__)(a, name=name, is_private=is_private, **k)
        self._session_ring = SessionRingModel(
            num_tracks,
            num_scenes,
            include_returns,
            set_session_highlight=set_session_highlight,
        )
        if tracks_to_use is not None:
            self._tracks_to_use = tracks_to_use
        else:
            self._tracks_to_use = partial(self._get_tracks_to_use, include_returns, include_master)
        if right_align_non_player_tracks:
            self._controlled_tracks_formatter = self._right_align_non_player_tracks
        else:
            self._controlled_tracks_formatter = self._pad_tracks
        self._cached_track_list = []
        self._update_track_list()
        self._snap_track_offset = snap_track_offset
        self.notify_offset(0, 0)
        if self.is_enabled():
            self._update_highlight()
        self.register_slot(self.song, self._update_track_list, "tracks")
        self.register_slot(self.song, self._update_track_list, "visible_tracks")
        self._SessionRingComponent__on_scene_list_changed.subject = self.song

    @property
    def num_tracks(self):
        return self._session_ring.num_tracks

    @property
    def num_scenes(self):
        return self._session_ring.num_scenes

    @property
    def track_offset(self):
        return self._session_ring.track_offset

    @track_offset.setter
    def track_offset(self, offset):
        self.set_offsets(offset, self.scene_offset)

    @property
    def scene_offset(self):
        return self._session_ring.scene_offset

    @scene_offset.setter
    def scene_offset(self, offset):
        self.set_offsets(self.track_offset, offset)

    @listenable_property
    def offset(self):
        return (self.track_offset, self.scene_offset)

    def tracks_to_use(self):
        return self._cached_track_list

    def scenes_to_use(self):
        return self.song.scenes

    @listenable_property
    def tracks(self):
        index = self.track_offset
        tracks = self.tracks_to_use()[index : index + self.num_tracks]
        return self._controlled_tracks_formatter(tracks)

    @listenable_property
    def scenes(self):
        index = self.scene_offset
        return self.scenes_to_use()[index : index + self.num_scenes]

    def set_offsets(self, track_offset, scene_offset):
        track_increment = 0
        scene_increment = 0
        if len(self.tracks_to_use()) > track_offset:
            track_increment = track_offset - self.track_offset
        if len(self.song.scenes) > scene_offset:
            scene_increment = scene_offset - self.scene_offset
        self.move(track_increment, scene_increment)

    def move(self, tracks, scenes):
        if self._snap_track_offset:
            tracks = tracks - tracks % self.num_tracks
        self._session_ring.move(tracks, scenes)
        self._update_highlight()
        self.notify_offset(self.track_offset, self.scene_offset)
        self.notify_tracks()
        self.notify_scenes()

    def on_enabled_changed(self):
        self._update_highlight()

    def _update_track_list(self):
        current_track_list = list(self._tracks_to_use())
        if self._cached_track_list != current_track_list:
            self._cached_track_list = current_track_list
            self.notify_tracks()
        new_offset = min(self.track_offset, len(current_track_list) - 1)
        if new_offset != self.track_offset:
            self.track_offset = new_offset

    def _update_highlight(self):
        if self.is_enabled() and self.num_tracks > 0 and self.num_scenes > 0:
            self._session_ring.update_highlight()
        else:
            self._session_ring.hide_highlight()

    def _pad_tracks(self, tracks):
        size = len(tracks)
        if size < self.num_tracks:
            tracks += [None] * (self.num_tracks - size)
        return tracks

    def _right_align_non_player_tracks(self, tracks):
        num_tracks = self.num_tracks
        insertion_index = index_if(lambda t: t not in self.song.tracks, tracks)
        tracks[insertion_index:insertion_index] = [None] * (num_tracks - len(tracks))
        return tracks

    def _get_tracks_to_use(self, include_returns, include_master):
        tracks = self.song.visible_tracks
        if include_returns:
            tracks += self.song.return_tracks
        if include_master:
            tracks += (self.song.master_track,)
        return tracks

    @listens("scenes")
    def __on_scene_list_changed(self):
        self.notify_scenes()
        new_offset = min(self.scene_offset, len(self.song.scenes) - 1)
        if new_offset != self.scene_offset:
            self.scene_offset = new_offset



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/_Framework/Layer.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 6736 bytes
from itertools import repeat

from future.utils import raise_

from .ControlElement import ControlElementClient
from .Disconnectable import Disconnectable
from .Resource import CompoundResource, ExclusiveResource
from .Util import nop


class LayerError(Exception):
    pass


class UnhandledControlError(LayerError):
    pass


class SimpleLayerOwner(Disconnectable):
    def __init__(self, layer=None):
        self._layer = layer
        self._layer.grab(self)

    def disconnect(self):
        self._layer.release(self)


class LayerClient(ControlElementClient):
    def __init__(self, layer=None, layer_client=None, *a, **k):
        (super().__init__)(*a, **k)
        self.layer_client = layer_client
        self.layer = layer

    def set_control_element(self, control_element, grabbed):
        layer = self.layer
        owner = self.layer_client
        names = layer._control_to_names[control_element]
        if not grabbed:
            control_element = None
        for name in names:
            try:
                handler = getattr(owner, "set_" + name)
            except AttributeError:
                try:
                    control = getattr(owner, name)
                    handler = control.set_control_element
                except AttributeError:
                    if name[0] != "_":
                        raise_(
                            UnhandledControlError,
                            f"Component {owner!s} has no handler for control_element {name}",
                        )
                    else:
                        handler = nop

            handler(control_element or None)
            layer._name_to_controls[name] = control_element


class LayerBase:
    pass


class CompoundLayer(LayerBase, CompoundResource):
    def _get_priority(self):
        return self.first.priority

    def _set_priority(self, priority):
        self.first.priority = priority
        self.second.priority = priority

    priority = property(_get_priority, _set_priority)

    def __getattr__(self, key):
        try:
            return getattr(self.first, key)
        except AttributeError:
            return getattr(self.second, key)


class Layer(LayerBase, ExclusiveResource):
    def __init__(self, priority=None, **controls):
        super().__init__()
        self._priority = priority
        self._name_to_controls = dict(zip(iter(controls.keys()), repeat(None)))
        self._control_to_names = {}
        self._control_clients = {}
        for name, control in controls.items():
            self._control_to_names.setdefault(control, []).append(name)

    def __add__(self, other):
        return CompoundLayer(self, other)

    def _get_priority(self):
        return self._priority

    def _set_priority(self, priority):
        if priority != self._priority:
            if self.owner:
                raise RuntimeError("Cannot change priority of a layer while it's owned")
            self._priority = priority

    priority = property(_get_priority, _set_priority)

    def __getattr__(self, name):
        try:
            return self._name_to_controls[name]
        except KeyError:
            raise AttributeError

    def grab(self, client, *a, **k):
        if client == self.owner:
            (self.on_received)(client, *a, **k)
            return True
        return (super().grab)(client, *a, **k)

    def on_received(self, client, *a, **k):
        for control in self._control_to_names:
            k.setdefault("priority", self._priority)
            (control.resource.grab)(self._get_control_client(client), *a, **k)

    def on_lost(self, client):
        for control in self._control_to_names:
            control.resource.release(self._get_control_client(client))

    def _get_control_client(self, client):
        try:
            control_client = self._control_clients[client]
        except KeyError:
            control_client = self._control_clients[client] = LayerClient(layer_client=client, layer=self)

        return control_client



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/_Generic/Devices.py
# Compiled at: 2023-12-21 15:35:35
# Size of source mod 2**32: 42603 bytes
from functools import partial

from _Framework.Util import group
from past.utils import old_div

RCK_BANK1 = ("Macro 1", "Macro 2", "Macro 3", "Macro 4", "Macro 5", "Macro 6", "Macro 7", "Macro 8")
RCK_BANK2 = ("Macro 9", "Macro 10", "Macro 11", "Macro 12", "Macro 13", "Macro 14", "Macro 15", "Macro 16")
RCK_BANKS = (RCK_BANK1, RCK_BANK2)
RCK_BOBS = (RCK_BANK1,)
RCK_BNK_NAMES = ("Macros", "Macros 2")
ALG_BANK1 = (
    "OSC1 Level",
    "OSC1 Octave",
    "OSC1 Semi",
    "OSC1 Shape",
    "OSC2 Level",
    "OSC2 Octave",
    "OSC2 Semi",
    "OSC2 Shape",
)
ALG_BANK2 = ("OSC1 Balance", "F1 Freq", "F1 Resonance", "F1 Type", "OSC2 Balance", "F2 Freq", "F2 Resonance", "F2 Type")
ALG_BANK3 = (
    "FEG1 Attack",
    "FEG1 Decay",
    "FEG1 Sustain",
    "FEG1 Rel",
    "FEG2 Attack",
    "FEG2 Decay",
    "FEG2 Sustain",
    "FEG2 Rel",
)
ALG_BANK4 = (
    "F1 On/Off",
    "F1 Freq < LFO",
    "F1 Freq < Env",
    "F1 Res < LFO",
    "F2 On/Off",
    "F2 Freq < LFO",
    "F2 Freq < Env",
    "F2 Res < LFO",
)
ALG_BANK5 = (
    "AEG1 Attack",
    "AEG1 Decay",
    "AEG1 Sustain",
    "AEG1 Rel",
    "AEG2 Attack",
    "AEG2 Decay",
    "AEG2 Sustain",
    "AEG2 Rel",
)
ALG_BANK6 = ("AMP1 Level", "AMP1 Pan", "LFO1 Shape", "LFO1 Speed", "AMP2 Level", "AMP2 Pan", "LFO2 Shape", "LFO2 Speed")
ALG_BANK7 = (
    "Volume",
    "Noise On/Off",
    "Noise Level",
    "Noise Color",
    "Unison On/Off",
    "Unison Detune",
    "Vib On/Off",
    "Vib Amount",
)
ALG_BOB = ("F1 Freq", "F1 Resonance", "OSC1 Shape", "OSC1 Octave", "OSC2 Shape", "OSC2 Octave", "OSC2 Detune", "Volume")
ALG_BANKS = (ALG_BANK1, ALG_BANK2, ALG_BANK3, ALG_BANK4, ALG_BANK5, ALG_BANK6, ALG_BANK7)
ALG_BOBS = (ALG_BOB,)
ALG_BNK_NAMES = ("Oscillators", "Filters", "Filter Envelopes", "Filter Modulation", "Volume Envelopes", "Mix", "Output")
COL_BANK1 = (
    "Mallet On/Off",
    "Mallet Volume",
    "Mallet Noise Amount",
    "Mallet Stiffness",
    "Mallet Noise Color",
    "",
    "",
    "",
)
COL_BANK2 = (
    "Noise Volume",
    "Noise Filter Type",
    "Noise Filter Freq",
    "Noise Filter Q",
    "Noise Attack",
    "Noise Decay",
    "Noise Sustain",
    "Noise Release",
)
COL_BANK3 = (
    "Res 1 Decay",
    "Res 1 Material",
    "Res 1 Type",
    "Res 1 Quality",
    "Res 1 Tune",
    "Res 1 Fine Tune",
    "Res 1 Pitch Env.",
    "Res 1 Pitch Env. Time",
)
COL_BANK4 = (
    "Res 1 Listening L",
    "Res 1 Listening R",
    "Res 1 Hit",
    "Res 1 Brightness",
    "Res 1 Inharmonics",
    "Res 1 Radius",
    "Res 1 Opening",
    "Res 1 Ratio",
)
COL_BANK5 = (
    "Res 2 Decay",
    "Res 2 Material",
    "Res 2 Type",
    "Res 2 Quality",
    "Res 2 Tune",
    "Res 2 Fine Tune",
    "Res 2 Pitch Env.",
    "Res 2 Pitch Env. Time",
)
COL_BANK6 = (
    "Res 2 Listening L",
    "Res 2 Listening R",
    "Res 2 Hit",
    "Res 2 Brightness",
    "Res 2 Inharmonics",
    "Res 2 Radius",
    "Res 2 Opening",
    "Res 2 Ratio",
)
COL_BOB = (
    "Res 1 Brightness",
    "Res 1 Type",
    "Mallet Stiffness",
    "Mallet Noise Amount",
    "Res 1 Inharmonics",
    "Res 1 Decay",
    "Res 1 Tune",
    "Volume",
)
COL_BANKS = (COL_BANK1, COL_BANK2, COL_BANK3, COL_BANK4, COL_BANK5, COL_BANK6)
COL_BOBS = (COL_BOB,)
COL_BNK_NAMES = (
    "Mallet",
    "Noise",
    "Resonator 1, Set A",
    "Resonator 1, Set B",
    "Resonator 2, Set A",
    "Resonator 2, Set B",
)
ELC_BANK1 = (
    "M Stiffness",
    "M Force",
    "Noise Pitch",
    "Noise Decay",
    "Noise Amount",
    "F Tine Color",
    "F Tine Decay",
    "F Tine Vol",
)
ELC_BANK2 = ("F Tone Decay", "F Tone Vol", "F Release", "Damp Tone", "Damp Balance", "Damp Amount", "", "")
ELC_BANK3 = ("P Symmetry", "P Distance", "P Amp In", "P Amp Out", "Pickup Model", "", "", "")
ELC_BANK4 = (
    "M Stiff < Vel",
    "M Stiff < Key",
    "M Force < Vel",
    "M Force < Key",
    "Noise < Key",
    "F Tine < Key",
    "P Amp < Key",
    "",
)
ELC_BANK5 = ("Volume", "Voices", "Semitone", "Detune", "KB Stretch", "PB Range", "Note PB Range", "")
ELC_BOB = ("M Stiffness", "M Force", "Noise Amount", "F Tine Vol", "F Tone Vol", "F Release", "P Symmetry", "Volume")
ELC_BANKS = (ELC_BANK1, ELC_BANK2, ELC_BANK3, ELC_BANK4, ELC_BANK5)
ELC_BOBS = (ELC_BOB,)
ELC_BNK_NAMES = ("Mallet and Tine", "Tone and Damper", "Pickup", "Modulation", "Global")
IMP_BANK1 = (
    "1 Start",
    "1 Transpose",
    "1 Stretch Factor",
    "1 Saturator Drive",
    "1 Filter Freq",
    "1 Filter Res",
    "1 Pan",
    "1 Volume",
)
IMP_BANK2 = (
    "2 Start",
    "2 Transpose",
    "2 Stretch Factor",
    "2 Saturator Drive",
    "2 Filter Freq",
    "2 Filter Res",
    "2 Pan",
    "2 Volume",
)
IMP_BANK3 = (
    "3 Start",
    "3 Transpose",
    "3 Stretch Factor",
    "3 Saturator Drive",
    "3 Filter Freq",
    "3 Filter Res",
    "3 Pan",
    "3 Volume",
)
IMP_BANK4 = (
    "4 Start",
    "4 Transpose",
    "4 Stretch Factor",
    "4 Saturator Drive",
    "4 Filter Freq",
    "4 Filter Res",
    "4 Pan",
    "4 Volume",
)
IMP_BANK5 = (
    "5 Start",
    "5 Transpose",
    "5 Stretch Factor",
    "5 Saturator Drive",
    "5 Filter Freq",
    "5 Filter Res",
    "5 Pan",
    "5 Volume",
)
IMP_BANK6 = (
    "6 Start",
    "6 Transpose",
    "6 Stretch Factor",
    "6 Saturator Drive",
    "6 Filter Freq",
    "6 Filter Res",
    "6 Pan",
    "6 Volume",
)
IMP_BANK7 = (
    "7 Start",
    "7 Transpose",
    "7 Stretch Factor",
    "7 Saturator Drive",
    "7 Filter Freq",
    "7 Filter Res",
    "7 Pan",
    "7 Volume",
)
IMP_BANK8 = (
    "8 Start",
    "8 Transpose",
    "8 Stretch Factor",
    "8 Saturator Drive",
    "8 Filter Freq",
    "8 Filter Res",
    "8 Pan",
    "8 Volume",
)
IMP_BOB = (
    "Global Time",
    "Global Transpose",
    "1 Transpose",
    "2 Transpose",
    "3 Transpose",
    "4 Transpose",
    "5 Transpose",
    "6 Transpose",
)
IMP_BANKS = (IMP_BANK1, IMP_BANK2, IMP_BANK3, IMP_BANK4, IMP_BANK5, IMP_BANK6, IMP_BANK7, IMP_BANK8)
IMP_BOBS = (IMP_BOB,)
IMP_BNK_NAMES = ("Pad 1", "Pad 2", "Pad 3", "Pad 4", "Pad 5", "Pad 6", "Pad 7", "Pad 8")
OPR_BANK1 = (
    "Ae Attack",
    "Ae Decay",
    "Ae Sustain",
    "Ae Release",
    "A Coarse",
    "A Fine",
    "Osc-A Lev < Vel",
    "Osc-A Level",
)
OPR_BANK2 = (
    "Be Attack",
    "Be Decay",
    "Be Sustain",
    "Be Release",
    "B Coarse",
    "B Fine",
    "Osc-B Lev < Vel",
    "Osc-B Level",
)
OPR_BANK3 = (
    "Ce Attack",
    "Ce Decay",
    "Ce Sustain",
    "Ce Release",
    "C Coarse",
    "C Fine",
    "Osc-C Lev < Vel",
    "Osc-C Level",
)
OPR_BANK4 = (
    "De Attack",
    "De Decay",
    "De Sustain",
    "De Release",
    "D Coarse",
    "D Fine",
    "Osc-D Lev < Vel",
    "Osc-D Level",
)
OPR_BANK5 = ("Le Attack", "Le Decay", "Le Sustain", "Le Release", "LFO Rate", "LFO Amt", "LFO Type", "LFO R < K")
OPR_BANK6 = (
    "Fe Attack",
    "Fe Decay",
    "Fe Sustain",
    "Fe Release",
    "Filter Freq",
    "Filter Res",
    "Fe R < Vel",
    "Fe Amount",
)
OPR_BANK7 = ("Pe Attack", "Pe Decay", "Pe Sustain", "Pe Release", "Pe Init", "Glide Time", "Pe Amount", "Spread")
OPR_BANK8 = ("Time < Key", "Panorama", "Pan < Key", "Pan < Rnd", "Algorithm", "Time", "Tone", "Volume")
OPR_BOB = ("Filter Freq", "Filter Res", "A Coarse", "A Fine", "B Coarse", "B Fine", "Osc-B Level", "Volume")
OPR_BANKS = (OPR_BANK1, OPR_BANK2, OPR_BANK3, OPR_BANK4, OPR_BANK5, OPR_BANK6, OPR_BANK7, OPR_BANK8)
OPR_BOBS = (OPR_BOB,)
OPR_BNK_NAMES = (
    "Oscillator A",
    "Oscillator B",
    "Oscillator C",
    "Oscillator D",
    "LFO",
    "Filter",
    "Pitch Modulation",
    "Routing",
)
SAM_BANK1 = ("Volume", "Ve Attack", "Ve Decay", "Ve Sustain", "Ve Release", "Vol < Vel", "Ve R < Vel", "Time")
SAM_BANK2 = (
    "Filter Type",
    "Filter Morph",
    "Filter Freq",
    "Filter Res",
    "Filt < Vel",
    "Filt < Key",
    "Fe < Env",
    "Shaper Amt",
)
SAM_BANK3 = ("Fe Attack", "Fe Decay", "Fe Sustain", "Fe Release", "Fe End", "Fe Mode", "Fe Loop", "Fe Retrig")
SAM_BANK4 = ("L 1 Wave", "L 1 Sync", "L 1 Sync Rate", "L 1 Rate", "Vol < LFO", "Filt < LFO", "Pan < LFO", "Pitch < LFO")
SAM_BANK5 = ("L 2 Wave", "L 2 Sync", "L 2 Sync Rate", "L 2 Rate", "L 2 R < Key", "L 2 St Mode", "L 2 Spin", "L 2 Phase")
SAM_BANK6 = ("L 3 Wave", "L 3 Sync", "L 3 Sync Rate", "L 3 Rate", "L 3 R < Key", "L 3 St Mode", "L 3 Spin", "L 3 Phase")
SAM_BANK7 = ("O Mode", "O Volume", "O Coarse", "O Fine", "Oe Attack", "Oe Decay", "Oe Sustain", "Oe Release")
SAM_BANK8 = ("Transpose", "Spread", "Pe < Env", "Pe Attack", "Pe Peak", "Pe Decay", "Pe Sustain", "Pe Release")
SAM_BOB = ("Filter Freq", "Filter Res", "Fe < Env", "Fe Decay", "Ve Attack", "Ve Release", "Transpose", "Volume")
SAM_BANKS = (SAM_BANK1, SAM_BANK2, SAM_BANK3, SAM_BANK4, SAM_BANK5, SAM_BANK6, SAM_BANK7, SAM_BANK8)
SAM_BOBS = (SAM_BOB,)
SAM_BNK_NAMES = ("Volume", "Filter", "Filter Envelope", "LFO 1", "LFO 2", "LFO 3", "Oscillator", "Pitch")
SIM_BANK1 = ("Ve Attack", "Ve Decay", "Ve Sustain", "Ve Release", "S Start", "S Loop Length", "S Length", "S Loop Fade")
SIM_BANK2 = ("Fe Attack", "Fe Decay", "Fe Sustain", "Fe Release", "Filter Freq", "Filter Res", "Filt < Vel", "Fe < Env")
SIM_BANK3 = ("L Attack", "L Rate", "L R < Key", "L Wave", "Vol < LFO", "Filt < LFO", "Pitch < LFO", "Pan < LFO")
SIM_BANK4 = ("Pe Attack", "Pe Decay", "Pe Sustain", "Pe Release", "Glide Time", "Spread", "Pan", "Volume")
SIM_BOB = ("Filter Freq", "Filter Res", "S Start", "S Length", "Ve Attack", "Ve Release", "Transpose", "Volume")
SIM_BANKS = (SIM_BANK1, SIM_BANK2, SIM_BANK3, SIM_BANK4)
SIM_BOBS = (SIM_BOB,)
SIM_BNK_NAMES = ("Amplitude", "Filter", "LFO", "Pitch Modifiers")
TNS_BANK1 = (
    "Exciter Type",
    "String Decay",
    "Str Inharmon",
    "Str Damping",
    "Exc ForceMassProt",
    "Exc FricStiff",
    "Exc Velocity",
    "E Pos",
)
TNS_BANK2 = (
    "Damper On",
    "Damper Mass",
    "D Stiffness",
    "D Velocity",
    "Damp Pos",
    "D Damping",
    "D Pos < Vel",
    "D Pos Abs",
)
TNS_BANK3 = (
    "Term On/Off",
    "Term Mass",
    "Term Fng Stiff",
    "Term Fret Stiff",
    "Pickup On/Off",
    "Pickup Pos",
    "T Mass < Vel",
    "T Mass < Key",
)
TNS_BANK4 = (
    "Body On/Off",
    "Body Type",
    "Body Size",
    "Body Decay",
    "Body Low-Cut",
    "Body High-Cut",
    "Body Mix",
    "Volume",
)
TNS_BANK5 = (
    "Vibrato On/Off",
    "Vib Delay",
    "Vib Fade-In",
    "Vib Speed",
    "Vib Amount",
    "Vib < ModWh",
    "Vib Error",
    "Volume",
)
TNS_BANK6 = (
    "Filter On/Off",
    "Filter Type",
    "Filter Freq",
    "Filter Reso",
    "Freq < Env",
    "Freq < LFO",
    "Reso < Env",
    "Reso < LFO",
)
TNS_BANK7 = (
    "FEG On/Off",
    "FEG Attack",
    "FEG Decay",
    "FEG Sustain",
    "FEG Release",
    "LFO On/Off",
    "LFO Shape",
    "LFO Speed",
)
TNS_BANK8 = ("Unison On/Off", "Uni Detune", "Porta On/Off", "Porta Time", "Voices", "Octave", "Semitone", "Volume")
TNS_BOB = (
    "Filter Freq",
    "Filter Reso",
    "Filter Type",
    "Exciter Type",
    "E Pos",
    "String Decay",
    "Str Damping",
    "Volume",
)
TNS_BANKS = (TNS_BANK1, TNS_BANK2, TNS_BANK3, TNS_BANK4, TNS_BANK5, TNS_BANK6, TNS_BANK7, TNS_BANK8)
TNS_BOBS = (TNS_BOB,)
TNS_BNK_NAMES = (
    "Exciter and String",
    "Damper",
    "Termination and Pickup",
    "Body",
    "Vibrato",
    "Filter",
    "Envelope and LFO",
    "Global",
)
WVT_BANK1 = (
    "Osc 1 Transp",
    "Osc 1 Detune",
    "Osc 1 Pos",
    "Osc 1 Effect 1",
    "Osc 1 Effect 2",
    "Osc 1 Pan",
    "Osc 1 Gain",
    "Osc 1 On",
)
WVT_BANK2 = (
    "Osc 2 Transp",
    "Osc 2 Detune",
    "Osc 2 Pos",
    "Osc 2 Effect 1",
    "Osc 2 Effect 2",
    "Osc 2 Pan",
    "Osc 2 Gain",
    "Osc 2 On",
)
WVT_BANK3 = (
    "Filter 1 On",
    "Filter 1 Freq",
    "Filter 1 Res",
    "Filter 1 Drive",
    "Filter 1 Type",
    "Filter 1 Slope",
    "Filter 1 LP/HP",
    "Filter 1 BP/NO/Morph",
)
WVT_BANK4 = (
    "Filter 2 On",
    "Filter 2 Freq",
    "Filter 2 Res",
    "Filter 2 Drive",
    "Filter 2 Type",
    "Filter 2 Slope",
    "Filter 2 LP/HP",
    "Filter 2 BP/NO/Morph",
)
WVT_BANK5 = (
    "Amp Attack",
    "Amp Decay",
    "Amp Sustain",
    "Amp Release",
    "Amp A Slope",
    "Amp D Slope",
    "Amp R Slope",
    "Amp Loop Mode",
)
WVT_BANK6 = (
    "Env 2 Attack",
    "Env 2 Decay",
    "Env 2 Sustain",
    "Env 2 Release",
    "Env 3 Attack",
    "Env 3 Decay",
    "Env 3 Sustain",
    "Env 3 Release",
)
WVT_BANK7 = (
    "LFO 1 Amount",
    "LFO 1 Rate",
    "LFO 1 S. Rate",
    "LFO 1 Sync",
    "LFO 2 Amount",
    "LFO 2 Rate",
    "LFO 2 S. Rate",
    "LFO 2 Sync",
)
WVT_BANK8 = ("Time", "Global Mod Amount", "Unison Amount", "Transpose", "Glide", "Sub Gain", "Sub Transpose", "Volume")
WVT_BOB = (
    "Osc 1 Pos",
    "Osc 1 Transp",
    "Osc 2 Pos",
    "Osc 2 Transp",
    "Filter 1 Freq",
    "Filter 1 Res",
    "Global Mod Amount",
    "Volume",
)
WVT_BANKS = (WVT_BANK1, WVT_BANK2, WVT_BANK3, WVT_BANK4, WVT_BANK5, WVT_BANK6, WVT_BANK7, WVT_BANK8)
WVT_BOBS = (WVT_BOB,)
WVT_BANK_NAMES = (
    "Oscillator 1",
    "Oscillator 2",
    "Filter 1",
    "Filter 2",
    "Amp Envelope",
    "Envelope 2/3",
    "LFO 1/2",
    "Global",
)
ARP_BANK1 = ("Style", "Groove", "Offset", "Synced Rate", "Retrigger Mode", "Ret. Interval", "Repeats", "Gate")
ARP_BANK2 = (
    "Tranpose Mode",
    "Tranpose Key",
    "Transp. Steps",
    "Transp. Dist.",
    "Velocity Decay",
    "Velocity Target",
    "Velocity On",
    "Vel. Retrigger",
)
ARP_BOB = (
    "Synced Rate",
    "Free Rate",
    "Transp. Steps",
    "Transp. Dist.",
    "Gate",
    "Tranpose Key",
    "Velocity Decay",
    "Velocity Target",
)
ARP_BANKS = (ARP_BANK1, ARP_BANK2)
ARP_BOBS = (ARP_BOB,)
ARP_BNK_NAMES = ("Style", "Pitch/Velocity")
CRD_BANK1 = ("Shift1", "Shift2", "Shift3", "Shift4", "Shift5", "Shift6", "", "")
CRD_BANK2 = ("Velocity1", "Velocity2", "Velocity3", "Velocity4", "Velocity5", "Velocity6", "", "")
CRD_BOB = ("Shift1", "Shift2", "Shift3", "Shift4", "Shift5", "Velocity5", "Shift6", "Velocity6")
CRD_BANKS = (CRD_BANK1, CRD_BANK2)
CRD_BOBS = (CRD_BOB,)
CRD_BNK_NAMES = ("Shift", "Shift %")
NTL_BANK1 = (
    "Sync On",
    "Time Length",
    "Synced Length",
    "Gate",
    "Release Velocity",
    "Decay Time",
    "Decay Key Scale",
    "Latch",
)
NTL_BANKS = (NTL_BANK1,)
NTL_BOBS = (NTL_BANK1,)
PIT_BANK1 = ("Pitch", "Range", "Lowest", "", "", "", "", "")
PIT_BANKS = (PIT_BANK1,)
PIT_BOBS = (PIT_BANK1,)
RND_BANK1 = ("Chance", "Choices", "Interval", "Sign", "", "", "", "")
RND_BANKS = (RND_BANK1,)
RND_BOBS = (RND_BANK1,)
SCL_BANK1 = ("Base", "Transpose", "Range", "Lowest", "", "", "", "")
SCL_BANKS = (SCL_BANK1,)
SCL_BOBS = (SCL_BANK1,)
VEL_BANK1 = ("Drive", "Compand", "Random", "Mode", "Out Hi", "Out Low", "Range", "Lowest")
VEL_BANKS = (VEL_BANK1,)
VEL_BOBS = (VEL_BANK1,)
AMP_BANK1 = ("Amp Type", "Bass", "Middle", "Treble", "Presence", "Gain", "Volume", "Dry/Wet")
AMP_BANK2 = ("Dual Mono", "", "", "", "", "", "", "")
AMP_BANKS = (AMP_BANK1, AMP_BANK2)
AMP_BOBS = (AMP_BANK1,)
AMP_BNK_NAMES = ("Global", "Dual Mono")
AFL_BANK1 = (
    "Frequency",
    "Resonance",
    "Env. Attack",
    "Env. Release",
    "Env. Modulation",
    "LFO Amount",
    "LFO Frequency",
    "LFO Phase",
)
AFL_BANK2 = (
    "Filter Type",
    "LFO Quantize On",
    "LFO Quantize Rate",
    "LFO Stereo Mode",
    "LFO Spin",
    "LFO Sync",
    "LFO Sync Rate",
    "LFO Offset",
)
AFL_BANK3 = ("", "", "", "", "", "S/C On", "S/C Mix", "S/C Gain")
AFL_BOB = (
    "Frequency",
    "Resonance",
    "Filter Type",
    "Env. Modulation",
    "LFO Amount",
    "LFO Waveform",
    "LFO Frequency",
    "LFO Phase",
)
AFL_BANKS = (AFL_BANK1, AFL_BANK2, AFL_BANK3)
AFL_BOBS = (AFL_BOB,)
AFL_BNK_NAMES = ("Filter", "Filter Extra", "Side Chain")
APN_BANK1 = ("Frequency", "Phase", "Shape", "Waveform", "Sync Rate", "Offset", "Width (Random)", "Amount")
APN_BANKS = (APN_BANK1,)
APN_BOBS = (APN_BANK1,)
BRP_BANK1 = ("Interval", "Offset", "Grid", "Variation", "Filter Freq", "Filter Width", "Volume", "Decay")
BRP_BANK2 = ("Chance", "Gate", "Pitch", "Pitch Decay", "Filter Freq", "Filter Width", "Volume", "Decay")
BRP_BOB = ("Grid", "Interval", "Offset", "Gate", "Pitch", "Pitch Decay", "Variation", "Chance")
BRP_BANKS = (BRP_BANK1, BRP_BANK2)
BRP_BOBS = (BRP_BOB,)
BRP_BNK_NAMES = ("Repeat Rate", "Gate/Pitch")
CAB_BANK1 = ("Cabinet Type", "Microphone Position", "Microphone Type", "Dual Mono", "", "", "", "Dry/Wet")
CAB_BANKS = (CAB_BANK1,)
CAB_BOBS = (CAB_BANK1,)
CHR_BANK1 = (
    "LFO Amount",
    "LFO Rate",
    "Delay 1 Time",
    "Delay 1 HiPass",
    "Delay 2 Time",
    "Delay 2 Mode",
    "Feedback",
    "Dry/Wet",
)
CHR_BANKS = (CHR_BANK1,)
CHR_BOBS = (CHR_BANK1,)
CP3_BANK1 = ("Threshold", "Ratio", "Attack", "Release", "Auto Release On/Off", "Env Mode", "Knee", "Model")
CP3_BANK2 = ("Threshold", "Expansion Ratio", "LookAhead", "S/C Listen", "S/C Gain", "Makeup", "Dry/Wet", "Output Gain")
CP3_BANK3 = ("S/C EQ On", "S/C EQ Type", "S/C EQ Freq", "S/C EQ Q", "S/C EQ Gain", "S/C On", "S/C Mix", "S/C Gain")
CP3_BOB = ("Threshold", "Ratio", "Attack", "Release", "Model", "Knee", "Dry/Wet", "Output Gain")
CP3_BANKS = (CP3_BANK1, CP3_BANK2, CP3_BANK3)
CP3_BOBS = (CP3_BOB,)
CP3_BNK_NAMES = ("Compression", "Output", "Side Chain")
CRP_BANK1 = ("Decay", "Material", "Mid Freq", "Width", "Bleed", "Resonance Type", "Gain", "Dry Wet")
CRP_BANK2 = ("Listening L", "Listening R", "Hit", "Brightness", "Inharmonics", "Radius", "Opening", "Ratio")
CRP_BANK3 = ("Resonance Type", "Tune", "Transpose", "Fine", "Spread", "Resonator Quality", "Note Off", "Off Decay")
CRP_BOB = ("Brightness", "Resonance Type", "Material", "Inharmonics", "Decay", "Ratio", "Tune", "Dry Wet")
CRP_BANKS = (CRP_BANK1, CRP_BANK2, CRP_BANK3)
CRP_BOBS = (CRP_BOB,)
CRP_BNK_NAMES = ("Amount", "Body", "Tune")
DRB_BANK1 = ("Drive", "Drive Type", "Transients", "Crunch", "Boom Freq", "Boom Amt", "Boom Decay", "Boom Audition")
DRB_BANK2 = ("Trim", "Output Gain", "Dry/Wet", "Compressor On", "Damping Freq", "", "", "")
DRB_BOB = ("Drive", "Drive Type", "Crunch", "Boom Amt", "Trim", "Damping Freq", "Output Gain", "Dry/Wet")
DRB_BANKS = (DRB_BANK1, DRB_BANK2)
DRB_BOBS = (DRB_BOB,)
DRB_BANK_NAMES = ("Drive", "Gain")
DTB_BANK1 = ("Drive", "Bias", "Envelope", "Tone", "Attack", "Release", "Output", "Dry/Wet")
DTB_BANKS = (DTB_BANK1,)
DTB_BOBS = (DTB_BANK1,)
ECH_BANK1 = ("L Division", "R Division", "L Sync Mode", "R Sync Mode", "L 16th", "R 16th", "L Sync", "R Sync")
ECH_BANK2 = ("L Time", "R Time", "L Offset", "R Offset", "Link", "Channel Mode", "", "")
ECH_BANK3 = ("Gate On", "Gate Thr", "Gate Release", "Duck On", "Duck Thr", "Duck Release", "Env Mix", "")
ECH_BANK4 = ("Noise On", "Noise Amt", "Noise Mrph", "Wobble On", "Wobble Amt", "Wobble Mrph", "Repitch", "")
ECH_BANK5 = ("Feedback", "Feedback Inv", "Input Gain", "Output Gain", "Clip Dry", "Dry Wet", "", "")
ECH_BANK6 = ("Filter On", "HP Freq", "HP Res", "LP Freq", "LP Res", "", "", "")
ECH_BANK7 = ("Reverb Level", "Reverb Decay", "Reverb Loc", "Stereo Width", "", "", "", "")
ECH_BANK8 = ("Mod Wave", "Mod Sync", "Mod Rate", "Mod Freq", "Mod Phase", "Dly < Mod", "Flt < Mod", "Mod 4x")
ECH_BOB = ("L Division", "R Division", "L Time", "R Time", "Input Gain", "Feedback", "Stereo Width", "Dry Wet")
ECH_BANKS = (ECH_BANK1, ECH_BANK2, ECH_BANK3, ECH_BANK4, ECH_BANK5, ECH_BANK6, ECH_BANK7, ECH_BANK8)
ECH_BOBS = (ECH_BOB,)
ECH_BANK_NAMES = ("Sync", "Time", "Gate/Ducking", "Noise/Wobble", "Gain", "Filter", "Reverb", "Modulation")
EQ8_BANK1 = (
    "1 Filter On A",
    "2 Filter On A",
    "3 Filter On A",
    "4 Filter On A",
    "5 Filter On A",
    "6 Filter On A",
    "7 Filter On A",
    "8 Filter On A",
)
EQ8_BANK2 = (
    "1 Frequency A",
    "2 Frequency A",
    "3 Frequency A",
    "4 Frequency A",
    "5 Frequency A",
    "6 Frequency A",
    "7 Frequency A",
    "8 Frequency A",
)
EQ8_BANK3 = ("1 Gain A", "2 Gain A", "3 Gain A", "4 Gain A", "5 Gain A", "6 Gain A", "7 Gain A", "8 Gain A")
EQ8_BANK4 = (
    "1 Resonance A",
    "2 Resonance A",
    "3 Resonance A",
    "4 Resonance A",
    "5 Resonance A",
    "6 Resonance A",
    "7 Resonance A",
    "8 Resonance A",
)
EQ8_BANK5 = (
    "1 Filter Type A",
    "2 Filter Type A",
    "3 Filter Type A",
    "4 Filter Type A",
    "5 Filter Type A",
    "6 Filter Type A",
    "7 Filter Type A",
    "8 Filter Type A",
)
EQ8_BANK6 = ("Adaptive Q", "", "", "", "", "", "Scale", "Output Gain")
EQ8_BANK7 = (
    "3 Gain A",
    "3 Frequency A",
    "3 Resonance A",
    "4 Gain A",
    "4 Frequency A",
    "4 Resonance A",
    "5 Gain A",
    "5 Frequency A",
)
EQ8_BOB = (
    "1 Frequency A",
    "1 Gain A",
    "2 Frequency A",
    "2 Gain A",
    "3 Frequency A",
    "3 Gain A",
    "4 Frequency A",
    "4 Gain A",
)
EQ8_BANKS = (EQ8_BANK1, EQ8_BANK2, EQ8_BANK3, EQ8_BANK4, EQ8_BANK5, EQ8_BANK6, EQ8_BANK7)
EQ8_BOBS = (EQ8_BOB,)
EQ8_BNK_NAMES = ("Band On/Off", "Frequency", "Gain", "Resonance", "Filter Type", "Output", "EQs 3-5")
EQ3_BANK1 = ("GainLo", "GainMid", "GainHi", "FreqLo", "LowOn", "MidOn", "HighOn", "FreqHi")
EQ3_BANKS = (EQ3_BANK1,)
EQ3_BOBS = (EQ3_BANK1,)
ERO_BANK1 = ("Frequency", "Width", "Mode", "Amount", "", "", "", "")
ERO_BANKS = (ERO_BANK1,)
ERO_BOBS = (ERO_BANK1,)
FLD_BANK1 = (
    "1 Filter Freq",
    "1 Filter Width",
    "1 Beat Delay",
    "1 Beat Swing",
    "1 Feedback",
    "1 Pan",
    "1 Volume",
    "Dry",
)
FLD_BANK2 = (
    "2 Filter Freq",
    "2 Filter Width",
    "2 Beat Delay",
    "2 Beat Swing",
    "2 Feedback",
    "2 Pan",
    "2 Volume",
    "Dry",
)
FLD_BANK3 = (
    "3 Filter Freq",
    "3 Filter Width",
    "3 Beat Delay",
    "3 Beat Swing",
    "3 Feedback",
    "3 Pan",
    "3 Volume",
    "Dry",
)
FLD_BOB = ("2 Filter Freq", "2 Filter Width", "2 Beat Delay", "2 Feedback", "1 Volume", "3 Volume", "2 Volume", "Dry")
FLD_BANKS = (FLD_BANK1, FLD_BANK2, FLD_BANK3)
FLD_BOBS = (FLD_BOB,)
FLD_BNK_NAMES = ("Input L Filter", "Input L+R Filter", "Input R Filter")
FLG_BANK1 = ("Hi Pass", "Dry/Wet", "Delay Time", "Feedback", "Env. Modulation", "Env. Attack", "Env. Release", "")
FLG_BANK2 = (
    "LFO Amount",
    "Frequency",
    "LFO Phase",
    "Sync",
    "LFO Offset",
    "Sync Rate",
    "LFO Width (Random)",
    "LFO Waveform",
)
FLG_BOB = ("Hi Pass", "Delay Time", "Frequency", "Sync Rate", "LFO Amount", "Env. Modulation", "Feedback", "Dry/Wet")
FLG_BANKS = (FLG_BANK1, FLG_BANK2)
FLG_BOBS = (FLG_BOB,)
FLG_BNK_NAMES = ("Frequency Controls", "LFO / S&H")
FRS_BANK1 = ("Coarse", "Fine", "Mode", "Ring Mod Frequency", "Drive On/Off", "Drive", "Wide", "Dry/Wet")
FRS_BANKS = (FRS_BANK1,)
FRS_BOBS = (FRS_BANK1,)
GTE_BANK1 = ("Threshold", "Return", "FlipMode", "LookAhead", "Attack", "Hold", "Release", "Floor")
GTE_BANK2 = ("S/C EQ On", "S/C EQ Type", "S/C EQ Freq", "S/C EQ Q", "S/C EQ Gain", "S/C On", "S/C Mix", "S/C Gain")
GTE_BANKS = (GTE_BANK1, GTE_BANK2)
GTE_BOBS = (GTE_BANK1,)
GTE_BNK_NAMES = ("Gate", "Side Chain")
GLU_BANK1 = ("Threshold", "Ratio", "Attack", "Release", "Peak Clip In", "Range", "Dry/Wet", "Makeup")
GLU_BANK2 = ("S/C EQ On", "S/C EQ Type", "S/C EQ Freq", "S/C EQ Q", "S/C EQ Gain", "S/C On", "S/C Mix", "S/C Gain")
GLU_BOB = ("Threshold", "Ratio", "Attack", "Release", "Peak Clip In", "Range", "Makeup", "Dry/Wet")
GLU_BANKS = (GLU_BANK1, GLU_BANK2)
GLU_BOBS = (GLU_BOB,)
GLU_BNK_NAMES = ("Compression", "Side Chain")
GRD_BANK1 = ("Frequency", "Pitch", "Time Delay", "Beat Swing", "Random", "Spray", "Feedback", "DryWet")
GRD_BANKS = (GRD_BANK1,)
GRD_BOBS = (GRD_BANK1,)
LPR_BANK1 = ("State", "Speed", "Reverse", "Quantization", "Monitor", "Song Control", "Tempo Control", "Feedback")
LPR_BANKS = (LPR_BANK1,)
LPR_BOBS = (LPR_BANK1,)
MBD_BANK1 = (
    "Master Output",
    "Amount",
    "Time Scaling",
    "Soft Knee On/Off",
    "Peak/RMS Mode",
    "Band Activator (High)",
    "Band Activator (Mid)",
    "Band Activator (Low)",
)
MBD_BANK2 = (
    "Input Gain (Low)",
    "Below Threshold (Low)",
    "Below Ratio (Low)",
    "Above Threshold (Low)",
    "Above Ratio (Low)",
    "Attack Time (Low)",
    "Release Time (Low)",
    "Output Gain (Low)",
)
MBD_BANK3 = (
    "Input Gain (Mid)",
    "Below Threshold (Mid)",
    "Below Ratio (Mid)",
    "Above Threshold (Mid)",
    "Above Ratio (Mid)",
    "Attack Time (Mid)",
    "Release Time (Mid)",
    "Output Gain (Mid)",
)
MBD_BANK4 = (
    "Input Gain (High)",
    "Below Threshold (High)",
    "Below Ratio (High)",
    "Above Threshold (High)",
    "Above Ratio (High)",
    "Attack Time (High)",
    "Release Time (High)",
    "Output Gain (High)",
)
MBD_BANK5 = ("Low-Mid Crossover", "Mid-High Crossover", "", "", "", "", "", "")
MBD_BANK6 = ("", "", "", "", "", "S/C On", "S/C Mix", "S/C Gain")
MBD_BOB = (
    "Above Threshold (Low)",
    "Above Ratio (Low)",
    "Above Threshold (Mid)",
    "Above Ratio (Mid)",
    "Above Threshold (High)",
    "Above Ratio (High)",
    "Master Output",
    "Amount",
)
MBD_BANKS = (MBD_BANK1, MBD_BANK2, MBD_BANK3, MBD_BANK4, MBD_BANK5, MBD_BANK6)
MBD_BOBS = (MBD_BOB,)
MBD_BNK_NAMES = ("Global", "Low Band", "Mid Band", "High Band", "Split Frequencies", "Side Chain")
OVR_BANK1 = ("Filter Freq", "Filter Width", "Drive", "Tone", "Preserve Dynamics", "", "", "Dry/Wet")
OVR_BANKS = (OVR_BANK1,)
OVR_BOBS = (OVR_BANK1,)
PDL_BANK1 = ("Type", "Gain", "Output", "Bass", "Mid", "Treble", "Sub", "Dry/Wet")
PDL_BANK2 = ("Mid Freq", "", "", "", "", "", "", "")
PDL_BANKS = (PDL_BANK1, PDL_BANK2)
PDL_BOBS = (PDL_BANK1,)
PDL_BANK_NAMES = ("General", "Eq")
PHS_BANK1 = ("Poles", "Color", "Dry/Wet", "Frequency", "Env. Modulation", "Env. Attack", "Env. Release", "Feedback")
PHS_BANK2 = (
    "LFO Amount",
    "LFO Frequency",
    "LFO Phase",
    "LFO Sync",
    "LFO Offset",
    "LFO Sync Rate",
    "LFO Spin",
    "LFO Waveform",
)
PHS_BOB = ("Frequency", "Feedback", "Poles", "Env. Modulation", "Color", "LFO Amount", "LFO Frequency", "Dry/Wet")
PHS_BANKS = (PHS_BANK1, PHS_BANK2)
PHS_BOBS = (PHS_BOB,)
PHS_BNK_NAMES = ("Frequency Controls", "LFO / S&H")
PPG_BANK1 = (
    "Filter Freq",
    "Filter Width",
    "Time Delay",
    "Beat Delay",
    "Beat Swing",
    "Delay Mode",
    "Feedback",
    "Dry/Wet",
)
PPG_BANKS = (PPG_BANK1,)
PPG_BOBS = (PPG_BANK1,)
RDX_BANK1 = ("Bit Depth", "Sample Mode", "Sample Hard", "Sample Soft", "Bit On", "", "", "")
RDX_BANKS = (RDX_BANK1,)
RDX_BOBS = (RDX_BANK1,)
RSN_BANK1 = ("Frequency", "Width", "Global Gain", "Dry/Wet", "Decay", "I Note", "Color", "I Gain")
RSN_BANK2 = ("II Gain", "III Gain", "IV Gain", "V Gain", "II Pitch", "III Pitch", "IV Pitch", "V Pitch")
RSN_BOB = ("Decay", "I Note", "II Pitch", "III Pitch", "IV Pitch", "V Pitch", "Global Gain", "Dry/Wet")
RSN_BANKS = (RSN_BANK1, RSN_BANK2)
RSN_BOBS = (RSN_BOB,)
RSN_BNK_NAMES = ("General / Mode I", "Modes II-IV")
RVB_BANK1 = (
    "In Filter Freq",
    "In Filter Width",
    "Predelay",
    "ER Spin On",
    "ER Spin Rate",
    "ER Spin Amount",
    "ER Shape",
    "Decay Time",
)
RVB_BANK2 = (
    "HiFilter Freq",
    "LowShelf Freq",
    "Chorus Rate",
    "Diffusion",
    "HiShelf Gain",
    "LowShelf Gain",
    "Chorus Amount",
    "Scale",
)
RVB_BANK3 = (
    "Decay Time",
    "Freeze On",
    "Room Size",
    "Stereo Image",
    "Reflect Level",
    "Diffuse Level",
    "Dry/Wet",
    "Density",
)
RVB_BOB = (
    "Decay Time",
    "Room Size",
    "Predelay",
    "In Filter Freq",
    "Reflect Level",
    "Diffuse Level",
    "Stereo Image",
    "Dry/Wet",
)
RVB_BANKS = (RVB_BANK1, RVB_BANK2, RVB_BANK3)
RVB_BOBS = (RVB_BOB,)
RVB_BNK_NAMES = ("Reflections", "Diffusion Network", "Global")
SAT_BANK1 = ("Drive", "Base", "Frequency", "Width", "Depth", "Output", "Dry/Wet", "Type")
SAT_BANK2 = ("WS Drive", "WS Lin", "WS Curve", "WS Damp", "WS Depth", "WS Period", "Dry/Wet", "")
SAT_BOB = ("Drive", "Type", "Base", "Frequency", "Width", "Depth", "Output", "Dry/Wet")
SAT_BANKS = (SAT_BANK1, SAT_BANK2)
SAT_BOBS = (SAT_BOB,)
SAT_BNK_NAMES = ("General Controls", "Waveshaper Controls")
SMD_BANK1 = (
    "L Beat Delay",
    "L Beat Swing",
    "L Time Delay",
    "R Beat Delay",
    "R Beat Swing",
    "R Time Delay",
    "Feedback",
    "Dry/Wet",
)
SMD_BANKS = (SMD_BANK1,)
SMD_BOBS = (SMD_BANK1,)
UTL_BANK1 = ("Left Inv", "Right Inv", "Channel Mode", "Stereo Width", "Mono", "Balance", "Gain", "Mute")
UTL_BANK2 = ("Bass Mono", "Bass Freq", "DC Filter", "", "", "", "", "")
UTL_BANKS = (UTL_BANK1, UTL_BANK2)
UTL_BOBS = (UTL_BANK1,)
UTL_BNK_NAMES = ("General Controls", "Low Frequency")
VDS_BANK1 = (
    "Tracing Freq.",
    "Tracing Width",
    "Tracing Drive",
    "Crackle Density",
    "Pinch Freq.",
    "Pinch Width",
    "Pinch Drive",
    "Crackle Volume",
)
VDS_BANKS = (VDS_BANK1,)
VDS_BOBS = (VDS_BANK1,)
VOC_BANK1 = (
    "Formant Shift",
    "Attack Time",
    "Release Time",
    "Mono/Stereo",
    "Output Level",
    "Gate Threshold",
    "Envelope Depth",
    "Dry/Wet",
)
VOC_BANK2 = (
    "Filter Bandwidth",
    "Upper Filter Band",
    "Lower Filter Band",
    "Precise/Retro",
    "Unvoiced Level",
    "Unvoiced Sensitivity",
    "Unvoiced Speed",
    "Enhance",
)
VOC_BANK3 = (
    "Noise Rate",
    "Noise Crackle",
    "Upper Pitch Detection",
    "Lower Pitch Detection",
    "Oscillator Pitch",
    "Oscillator Waveform",
    "Ext. In Gain",
    "",
)
VOC_BOB = (
    "Formant Shift",
    "Attack Time",
    "Release Time",
    "Unvoiced Level",
    "Gate Threshold",
    "Filter Bandwidth",
    "Envelope Depth",
    "Dry/Wet",
)
VOC_BANKS = (VOC_BANK1, VOC_BANK2, VOC_BANK3)
VOC_BOBS = (VOC_BOB,)
VOC_BNK_NAMES = ("Global", "Filters/Voicing", "Carrier")
DEVICE_DICT = {
    "AudioEffectGroupDevice": RCK_BANKS,
    "MidiEffectGroupDevice": RCK_BANKS,
    "InstrumentGroupDevice": RCK_BANKS,
    "DrumGroupDevice": RCK_BANKS,
    "InstrumentImpulse": IMP_BANKS,
    "Operator": OPR_BANKS,
    "UltraAnalog": ALG_BANKS,
    "OriginalSimpler": SIM_BANKS,
    "MultiSampler": SAM_BANKS,
    "MidiArpeggiator": ARP_BANKS,
    "LoungeLizard": ELC_BANKS,
    "StringStudio": TNS_BANKS,
    "Collision": COL_BANKS,
    "MidiChord": CRD_BANKS,
    "MidiNoteLength": NTL_BANKS,
    "MidiPitcher": PIT_BANKS,
    "MidiRandom": RND_BANKS,
    "MidiScale": SCL_BANKS,
    "MidiVelocity": VEL_BANKS,
    "AutoFilter": AFL_BANKS,
    "AutoPan": APN_BANKS,
    "BeatRepeat": BRP_BANKS,
    "Chorus": CHR_BANKS,
    "Compressor2": CP3_BANKS,
    "Corpus": CRP_BANKS,
    "Eq8": EQ8_BANKS,
    "FilterEQ3": EQ3_BANKS,
    "Erosion": ERO_BANKS,
    "FilterDelay": FLD_BANKS,
    "Flanger": FLG_BANKS,
    "FrequencyShifter": FRS_BANKS,
    "GrainDelay": GRD_BANKS,
    "Looper": LPR_BANKS,
    "MultibandDynamics": MBD_BANKS,
    "Overdrive": OVR_BANKS,
    "Phaser": PHS_BANKS,
    "Redux": RDX_BANKS,
    "Saturator": SAT_BANKS,
    "Resonator": RSN_BANKS,
    "StereoGain": UTL_BANKS,
    "Tube": DTB_BANKS,
    "Reverb": RVB_BANKS,
    "Vinyl": VDS_BANKS,
    "Gate": GTE_BANKS,
    "Vocoder": VOC_BANKS,
    "Amp": AMP_BANKS,
    "Cabinet": CAB_BANKS,
    "GlueCompressor": GLU_BANKS,
    "Pedal": PDL_BANKS,
    "DrumBuss": DRB_BANKS,
    "Echo": ECH_BANKS,
    "InstrumentVector": WVT_BANKS,
}
DEVICE_BOB_DICT = {
    "AudioEffectGroupDevice": RCK_BOBS,
    "MidiEffectGroupDevice": RCK_BOBS,
    "InstrumentGroupDevice": RCK_BOBS,
    "DrumGroupDevice": RCK_BOBS,
    "InstrumentImpulse": IMP_BOBS,
    "Operator": OPR_BOBS,
    "UltraAnalog": ALG_BOBS,
    "OriginalSimpler": SIM_BOBS,
    "MultiSampler": SAM_BOBS,
    "MidiArpeggiator": ARP_BOBS,
    "LoungeLizard": ELC_BOBS,
    "StringStudio": TNS_BOBS,
    "Collision": COL_BOBS,
    "MidiChord": CRD_BOBS,
    "MidiNoteLength": NTL_BOBS,
    "MidiPitcher": PIT_BOBS,
    "MidiRandom": RND_BOBS,
    "MidiScale": SCL_BOBS,
    "MidiVelocity": VEL_BOBS,
    "AutoFilter": AFL_BOBS,
    "AutoPan": APN_BOBS,
    "BeatRepeat": BRP_BOBS,
    "Chorus": CHR_BOBS,
    "Compressor2": CP3_BOBS,
    "Corpus": CRP_BOBS,
    "Eq8": EQ8_BOBS,
    "FilterEQ3": EQ3_BOBS,
    "Erosion": ERO_BOBS,
    "FilterDelay": FLD_BOBS,
    "Flanger": FLG_BOBS,
    "FrequencyShifter": FRS_BOBS,
    "GrainDelay": GRD_BOBS,
    "Looper": LPR_BOBS,
    "MultibandDynamics": MBD_BOBS,
    "Overdrive": OVR_BOBS,
    "Phaser": PHS_BOBS,
    "Redux": RDX_BOBS,
    "Saturator": SAT_BOBS,
    "Resonator": RSN_BOBS,
    "StereoGain": UTL_BOBS,
    "Tube": DTB_BOBS,
    "Reverb": RVB_BOBS,
    "Vinyl": VDS_BOBS,
    "Gate": GTE_BOBS,
    "Vocoder": VOC_BOBS,
    "Amp": AMP_BOBS,
    "Cabinet": CAB_BOBS,
    "GlueCompressor": GLU_BOBS,
    "Pedal": PDL_BOBS,
    "DrumBuss": DRB_BOBS,
    "Echo": ECH_BOBS,
    "InstrumentVector": WVT_BOBS,
}
BANK_NAME_DICT = {
    "AudioEffectGroupDevice": RCK_BNK_NAMES,
    "MidiEffectGroupDevice": RCK_BNK_NAMES,
    "InstrumentGroupDevice": RCK_BNK_NAMES,
    "DrumGroupDevice": RCK_BNK_NAMES,
    "InstrumentImpulse": IMP_BNK_NAMES,
    "Operator": OPR_BNK_NAMES,
    "UltraAnalog": ALG_BNK_NAMES,
    "OriginalSimpler": SIM_BNK_NAMES,
    "MultiSampler": SAM_BNK_NAMES,
    "MidiArpeggiator": ARP_BNK_NAMES,
    "LoungeLizard": ELC_BNK_NAMES,
    "StringStudio": TNS_BNK_NAMES,
    "Collision": COL_BNK_NAMES,
    "MidiChord": CRD_BNK_NAMES,
    "BeatRepeat": BRP_BNK_NAMES,
    "Compressor2": CP3_BNK_NAMES,
    "Corpus": CRP_BNK_NAMES,
    "Eq8": EQ8_BNK_NAMES,
    "FilterDelay": FLD_BNK_NAMES,
    "Flanger": FLG_BNK_NAMES,
    "Gate": GTE_BNK_NAMES,
    "MultibandDynamics": MBD_BNK_NAMES,
    "Phaser": PHS_BNK_NAMES,
    "Saturator": SAT_BNK_NAMES,
    "Resonator": RSN_BNK_NAMES,
    "Reverb": RVB_BNK_NAMES,
    "Vocoder": VOC_BNK_NAMES,
    "Amp": AMP_BNK_NAMES,
    "GlueCompressor": GLU_BNK_NAMES,
    "AutoFilter": AFL_BNK_NAMES,
    "StereoGain": UTL_BNK_NAMES,
    "DrumBuss": DRB_BANK_NAMES,
    "Echo": ECH_BANK_NAMES,
    "Pedal": PDL_BANK_NAMES,
    "InstrumentVector": WVT_BANK_NAMES,
}
MAX_DEVICES = ("MxDeviceInstrument", "MxDeviceAudioEffect", "MxDeviceMidiEffect")


def device_parameters_to_map(device):
    return tuple(device.parameters[1:])


def parameter_bank_names(device, bank_name_dict=BANK_NAME_DICT):
    if device is not None:
        if device.class_name in list(bank_name_dict.keys()):
            return bank_name_dict[device.class_name]
        banks = number_of_parameter_banks(device)

        def _default_bank_name(bank_index):
            return "Bank " + str(bank_index + 1)

        if device.class_name in MAX_DEVICES:
            if banks != 0:

                def _is_ascii(c):
                    return ord(c) < 128

                def _bank_name(bank_index):
                    try:
                        name = device.get_bank_name(bank_index)
                    except:
                        name = None

                    if name:
                        return "".join(filter(_is_ascii, name))
                    return _default_bank_name(bank_index)

                return list(map(_bank_name, list(range(banks))))
        return list(map(_default_bank_name, list(range(banks))))
    return []


def parameter_banks(device, device_dict=DEVICE_DICT):
    if device is not None:
        if device.class_name in list(device_dict.keys()):

            def names_to_params(bank):
                return list(map(partial(get_parameter_by_name, device), bank))

            return list(map(names_to_params, device_dict[device.class_name]))
        if device.class_name in MAX_DEVICES:
            try:
                banks = device.get_bank_count()
            except:
                banks = 0

            if banks != 0:

                def _bank_parameters(bank_index):
                    try:
                        parameter_indices = device.get_bank_parameters(bank_index)
                    except:
                        parameter_indices = []

                    if len(parameter_indices) != 8:
                        return [None for i in range(8)]
                    return [device.parameters[i] if i != -1 else None for i in parameter_indices]

                return list(map(_bank_parameters, list(range(banks))))
        return group(device_parameters_to_map(device), 8)
    return []


def best_of_parameter_bank(device, device_bob_dict=DEVICE_BOB_DICT):
    if device:
        if device.class_name in device_bob_dict:
            bobs = device_bob_dict[device.class_name]
            return list(map(partial(get_parameter_by_name, device), bobs[0]))
    if device.class_name in MAX_DEVICES:
        try:
            parameter_indices = device.get_bank_parameters(-1)
            return [device.parameters[i] if i != -1 else None for i in parameter_indices]
        except:
            return []

        return device.parameters[1:9]
    return None


def number_of_parameter_banks(device, device_dict=DEVICE_DICT):
    if device is not None:
        if device.class_name in list(device_dict.keys()):
            device_bank = device_dict[device.class_name]
            return len(device_bank)
        if device.class_name in MAX_DEVICES:
            try:
                banks = device.get_bank_count()
            except:
                banks = 0

            if banks != 0:
                return banks
        param_count = len(device.parameters[1:])
        return old_div(param_count, 8) + (1 if param_count % 8 else 0)
    return 0


def get_parameter_by_name(device, name):
    for i in device.parameters:
        if i.original_name == name:
            return i
    return None
